# Copyright (C) 2024- Davide Mollica <davide.mollica@inaf.it>
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of iactsim.
#
# iactsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# iactsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iactsim.  If not, see <https://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod
import random
import time
from typing import List, Union, Tuple

import cupy as cp
import numpy as np

from ._waveforms import Waveform
from._camera_geometry import ModularSiPMCameraGeometry
from ._pde import PhotonDetectionEfficiency

from .signals.sipm_signals._sipm_kernels import (
    sipm_signals,
    sipm_signals_w_ucells,
)

from .signals.discriminator_signals._disciminator_kernels import (
    ideal_discriminator,
)

from .signals.sampling_signals._sampling_kernels import (
    peak_detection,
)

from ..utils._timer import BenchmarkTimer

class CherenkovCamera(ABC):
    """Abstract class to simulate a Cherenkov camera. 

    Parameters
    ----------
    n_pixels : int
        Number of pixels.
    waveforms : Union[List[Waveform], Tuple[Waveform]]
        List of `Waveform` instances, one for each channel.
    trigger_channels : List[int]
        Trigger channels index.
    sampling_channels : List[int]
        Sampling channels index.
    channels_time_resolution : List[float]
        Time resolution of each channel.

    Notes
    -----
    **Random Seed Handling**

    The camera manages random number generation for stochastic processes (e.g.,
    cross-talk, background noise) on a per-pixel basis. The seed for each pixel
    is derived from a single ``uint64`` master seed, :py:attr:`base_seed`, to
    ensure reproducibility. This behavior is controlled by the following
    attributes:

    - :py:attr:`base_seed`: The master seed that can be set by the user for
      reproducibility. When this value is changed, the per-pixel seeds in
      :py:attr:`seed` are automatically updated.
    - :py:attr:`random_seed`: If ``True`` (default), a new random value for
      :py:attr:`base_seed` is generated at each :py:meth:`restart()` call.
      This ensures that different simulation runs are independent. When set to ``False``,
      :py:attr:`base_seed` is reset to 0.
    - :py:attr:`increment_seed`: If ``True`` (default), the :py:attr:`base_seed`
      is incremented by `n_pixels` after each :py:meth:`simulate_response()`
      call. This provides different random sequences for consecutive events
      within the same run.
    - :py:attr:`seed`: A CuPy array containing a unique seed for each pixel,
      derived from :py:attr:`base_seed`. This array is passed to the CUDA
      kernels.


    """
    def __init__(
        self,
        n_pixels: int,
        waveforms: Tuple[Waveform],
        trigger_channels: List[int],
        sampling_channels: List[int],
        channels_time_resolution: List[float],
    ):
        self.n_pixels = n_pixels #: Number of pixels.
        self.pixel_mask = cp.zeros((self.n_pixels,), dtype=cp.int32) #: Pixel mask. Pixel with value 1 are ignored.

        # Channels waveform
        self.n_channels = None #: Number of channels.
        self.waveforms = waveforms #: Single photo-electron waveform of each channel.

        # Split channels (to compute trigger signals first)
        self.trigger_channels = trigger_channels #: Trigger channels list.
        self.sampling_channels = sampling_channels #: Sampling channels list.

        self.auxiliary_channels = [] #: Auxiliary channels list.
        for i in range(self.n_channels):
            if i not in trigger_channels + sampling_channels:
                self.auxiliary_channels.append(i)
        
        # Channels time window
        self.time_windows = [None]*self.n_channels #: Time windows where to simulate the signal. One for each channel.
        self.fixed_time_windows = False #: Whether time windows should be computed for each event or assumed fixed.
        self.channels_time_resolution = channels_time_resolution #: Time resolution for each channel.

        # Modulate the amplitude of the waveforms for each pixel
        self.channels_gain = [cp.ones((self.n_pixels,), dtype=cp.float32) for _ in self.waveforms] #: Gain of each channel.

        # Input photo-electrons (discharges initiated by photons)
        self.source = None #: Input photo-electrons arrival times.

        # Signals
        self.signals = cp.asarray([]) #: Computed signals.

        # Dynamic time windows
        # (used if self.fixed_time_windows is False)
        # for every channel in self.sampling_channels
        self.sampling_delay = [0.]*len(self.sampling_channels) #: Delay of the sampling time windows with respect to the camera trigger. One for each sampling channel.
        self.sampling_window_extent = [0.]*len(self.sampling_channels) #: Sampling windows extent. One value for each sampling channel.
        self.auxiliary_delay = [0.]*len(self.auxiliary_channels) #: Delay of the auxiliary time windows with respect to the camera trigger. One for each auxiliary channel.
        self.auxiliary_window_extent = [0.]*len(self.auxiliary_channels) #: Auxiliary windows extent. One value for each auxiliary channel.
        self.trigger_window_start_offset = [0.]*len(self.trigger_channels) #: Trigger channels window start offset with respect to the first photo-electron arrival time in :py:attr:`source`.
        self.trigger_window_end_offset = [0.]*len(self.trigger_channels) #: Trigger channels window end offset with respect to the last photo-electron arrival time in :py:attr:`source`.

        # Trigger info
        self.enable_camera_trigger = True #: Whether to enable the camera trigger or skip directly to the sampling phase (in this case a trigger time must be provided).
        self.triggered = False #: Whether the camera has been triggered.
        self.trigger_time = None #: Time of the last trigger.

        # Seed
        self.seed = None #: Seed state, must be different for each pixel.
        self.increment_seed = True 
        """Whether to increment the seed at each :py:meth:`simulate_response()` call."""
        self.random_seed = True

        # Benchmark
        self.timer = BenchmarkTimer() #: Benchmark timer instance.
        self.timer.add_section('simulate_response')

        # Event counter
        self._event_counter = 0 #: Number of registered events.
        self._triggered_events = [] #: List of triggered events.

    def restart(self):
        self._event_counter = 0
        self._triggered_events = []

    @property
    def random_seed(self):
        """Whether to set a random seed state at each :py:meth:`restart()` call."""
        return self._random_seed

    @random_seed.setter
    def random_seed(self, value: bool):
        self._random_seed = value
        if value:
            self.base_seed = random.getrandbits(63)
        else:
            self.base_seed = 0
    
    def _update_seed(self):
        """Create a unique seed for each pixel."""
        self.seed = cp.uint64(self.base_seed) + cp.arange(0, self.n_pixels, 1, dtype=cp.uint64)
    
    @property
    def base_seed(self):
        """Run-wise seed from wich per-pixel seeds are generated."""
        return self._base_seed
    
    @base_seed.setter
    def base_seed(self, value):
        self._base_seed = value
        self._update_seed()

    @property
    def triggered_events(self):
        return np.asarray(self._triggered_events, dtype=np.int32)

    @property
    def source(self):
        """Tuple containing:

              * *pes*: ndarray with shape (n_pe,) of photo-electron arrival times;
              * *mapping*: ndarray with shape (n_pixels+1,) of first and last arrival time position inside *pes* for each pixel.
            
            For instance:

              * discharge times on pixel 0 -> ``pes[mapping[0]:mapping[1]]``
              * discharge times on pixel n -> ``pes[mapping[n]:mapping[n+1]]``

            If None simulate only background.
        """
        return self._source

    @source.setter
    def source(self, a_source):
        if a_source is not None:
            t0s = cp.asarray(a_source[0], dtype=cp.float32)
            map_ = cp.asarray(a_source[1], dtype=cp.int32)
            if getattr(self, 'simulate_ucells', False):
                ucells_ids = cp.asarray(a_source[2], dtype=cp.int32)
                a_source = (t0s, map_, ucells_ids)
            else:
                a_source = (t0s, map_)
        else:
            if getattr(self, 'simulate_ucells', False):
                a_source = (cp.asarray([], dtype=cp.float32), cp.zeros((self.n_pixels+1), dtype=cp.int32), cp.asarray([], dtype=cp.int32))
            else:
                a_source = (cp.asarray([], dtype=cp.float32), cp.zeros((self.n_pixels+1), dtype=cp.int32))
        self._source = a_source

    @property
    def waveforms(self):
        """List of `Waveform` instances, one for each channel.
        """
        return self._waveforms
    
    @waveforms.setter
    def waveforms(self, waves):
        waveforms_are_ok = False
        
        # Waveforms list must be immutable
        if isinstance(waves, list):
            waves = tuple(waves)
        
        if isinstance(waves, tuple):
            waveforms_are_ok = all([isinstance(wf, Waveform) for wf in waves])
        
        if not waveforms_are_ok:
            raise(ValueError("Waveforms must be provided as a tuple of `Waveform` instances."))

        if self.n_channels is None:
            self.n_channels = len(waves)
        else:
            if len(waves) != self.n_channels:
                raise(RuntimeError(f'The camera has been initialized with {self.n_channels} default waveforms, but you are providing {len(waves)} waveforms.'))
        self._waveforms = waves
        
        # In this way it is possible to set waveforms attribute later
        # in a custom configuration
        if any([w.amplitude is None for w in waves]):
            self._waveforms_are_initialized = False
            return

        self._waveforms_are_initialized = True
        
        self._max_wave_extent = max([wf.get_extent() for wf in self._waveforms])

        start_time = []
        texture_pointer = []
        inv_dt = []
        self._texture_waveform_objects = []
        for wave in self.waveforms:
            obj, inv_dx, start = wave.get_waveform_texture()
            self._texture_waveform_objects.append(obj)
            texture_pointer.append(obj.ptr)
            inv_dt.append(inv_dx)
            start_time.append(start)

        start_time = cp.asarray(start_time, dtype=cp.float32)
        inv_dt = cp.asarray(inv_dt, dtype=cp.float32)
        tex_handles_gpu = cp.asarray(texture_pointer, dtype=cp.uint64)

        self._waveforms_params = [
            tex_handles_gpu,
            inv_dt,
            start_time
        ]

    def get_channel_signals(self, channel, reshape=False):
        """Retrieves the signals for a specified channel.

        Parameters
        ----------
        channel : int
            The channel number for which to retrieve signals.
        reshape : bool, optional
            If True, reshapes the output array to have dimensions (n_pixels, number of time windows).
            If False, returns a 1D array. Default is False.

        Returns
        -------
        cp.ndarray
            A CuPy array containing the signals for the specified channel.  The shape of the array
            depends on the `reshape` parameter.

        Raises
        ------
        RuntimeError
            If the time window has not been defined for the specified channel.
        """
        if channel > self.n_channels or self.time_windows[channel] is None:
            raise(RuntimeError(f'Time window have not been defined for channel {channel}.'))
        windows_map = np.cumsum([0]+[len(w) for w in self.time_windows])
        signals_map = windows_map*self.n_pixels
        signals = self.signals[signals_map[channel]:signals_map[channel+1]]
        if reshape:
            signals = signals.reshape(self.n_pixels,windows_map[channel+1]-windows_map[channel])
        return signals

    def _get_time_windows_parameters(self):
        windows = np.concatenate(self.time_windows)
        windows_map = np.cumsum([0]+[len(w) for w in self.time_windows], dtype=np.int32)
        return [
            cp.asarray(windows, cp.float32),
            windows_map,
            cp.asarray(windows_map, cp.int32)
        ]

    @abstractmethod
    def compute_signals(self, channels=None):
        """Method to compute signals of each pixel for the desired channels.

        Parameters
        ----------
        channels : list of channel indices
            If None compute signals of all channels. By default None.
        
        Notes
        -----
        The specific computing logic is left to the concrete implementations.
        However, implementations should ensure that the signal is only computed 
        for the specified channels and stored in the `signals` attribute array.
        """
        pass

    @abstractmethod
    def pre_trigger_action(self):
        """Computes trigger signals.

        Notes
        -----
        The specific logic is left to the concrete implementations.
        However, implementations should ensure that the signal is only computed 
        for the channels needed to generate a camera trigger.
        This method is called only if the camera trigger is enabled.
        """
        pass

    @abstractmethod
    def trigger_action(self):
        """Generates a camera trigger.

        Notes
        -----
        Concrete implementations of this method should contain the logic
        to determine when a camera trigger has been be generated.  If the
        trigger condition is met, implementations *must* ensure that:

          * the `triggered` attribute is set to `True`.
          * the `trigger_time` attribute is set to the time
            representing the moment when the trigger condition was met.
        
        This method is called only if the camera trigger is enabled.
        """
        pass

    def post_trigger_action(self):
        """Optional action that is performed right after the trigger action,
        regardless if a camera trigger is generated.
        This method is called only if the camera trigger is enabled.
        """
        pass

    @abstractmethod
    def pre_sampling_action(self):
        """Computes signals to be sampled.

        Notes
        -----
        The specific logic is left to the concrete implementations.
        However, implementations should ensure that the signals are only computed 
        for the channels that will be sampled (unless they are the same from which
        the camera trigger is generated).
        This method is called only if a trigger condition is met or if 
        the camera trigger is disabled.
        """
        pass

    @abstractmethod
    def sampling_action(self):
        """Digitise the signals.

        Notes
        -----
        The specific logic is left to the concrete implementations.
        This method is called only if a trigger condition is met or if 
        the camera trigger is disabled.
        """
        pass

    def post_sampling_action(self):
        """Optional action that is performed right after the sampling action.
        This method is called only if a trigger condition is met or if 
        the camera trigger is disabled.
        """
        pass

    def apply_ideal_discriminator_to_channel(
        self,
        channel: int,
        thresholds: cp.ndarray,
        offset: cp.ndarray,
        time_slices_over_threshold: int
    ) -> cp.ndarray:
        """Apply an ideal discriminator to a specific channel pre-computed signals.

        This method implements an ideal discriminator on the waveforms of a given
        channel.  It compares the input signal against per-pixel thresholds and
        sets the output signal based on whether the input signal exceeds the
        threshold for at least a specified number of consecutive time slices.

        Parameters
        ----------
        channel : int
            The channel number to process.
        thresholds : cp.ndarray
            A CuPy 1D array of thresholds, one for each pixel. 
            This array should have a dtype of `cp.float32`.
        offset : cp.ndarray
            A CuPy 1D array of threshold offsets, one for each pixel. 
            This array should have a dtype of `cp.float32`.
        time_slices_over_threshold : int
            The minimum number of consecutive time slices the signal must be 
            above the threshold to trigger a positive output. This value is 
            assumed for all pixels.

        Returns
        -------
        cp.ndarray
            A CuPy array of the same shape as the input signal, containing the
            output of the ideal discriminator. The output will be 0.0f where the
            signal is below the threshold (or doesn't stay above threshold long
            enough) and 1.0f where it is above the threshold.

        Examples
        --------
        .. code-block:: python

            n_pixels = camera.n_pixels
            thresholds = cp.random.rand(n_pixels).astype(cp.float32) * 10
            offset = cp.zeros((n_pixels), dtype=cp.float32)
            min_time_slices = 10
            channel_num = 0

            output_signal = camera.apply_ideal_discriminator_to_channel(channel_num, thresholds, offset, min_time_slices)
            # output_signal now contains the discriminated signal for channel 0.
        
        """

        inp_signal = self.get_channel_signals(channel, reshape=False)
        out_signal = cp.empty_like(inp_signal)
        window_size = len(self.time_windows[channel])
        args = [
            inp_signal,
            out_signal,
            thresholds,
            offset,
            cp.int32(time_slices_over_threshold),
            self.pixel_mask,
            cp.int32(window_size),
            cp.int32(self.n_pixels)
        ]
        
        block_size = self._n_threads

        shered_memory_size = window_size * 4
        
        ideal_discriminator(
            (self.n_pixels,),
            (block_size,),
            args,
            shared_mem = shered_memory_size
        )
        return out_signal

    def apply_peak_detection_to_channel(
        self,
        peak_amplitudes: cp.ndarray,
        peaking_times: cp.ndarray,
        channel: int,
        t_start: float,
        extent: float
    ) -> None:
        """Apply peak detection to a specific channel.

        This method processes the pre-computed signals for a given channel using `peak_detection` kernel
        to identify peaks and store their amplitudes and times.

        Parameters
        ----------
        peak_amplitudes : cp.ndarray
            A CuPy array where the detected peak amplitudes will be stored.
            This array should be pre-allocated with a size equal to the number 
            of pixels and have a dtype of `cp.float32`.
            It can be allocated once as class attribute.
        peaking_times : cp.ndarray
            A CuPy array where the detected peaking times will be stored.
            This array should be pre-allocated with a size equal to the number 
            of pixels and have a dtype of `cp.float32`.
            It can be allocated once as class attribute.
        channel : int
            The channel number to process.
        t_start : float
            The starting time of the peak detection time window (in ns).
        extent : float
            The duration of the peak detection time window (in ns).

        Examples
        --------

        .. code-block:: python

            n_pixels = camera.n_pixels
            peak_amps = cp.empty((n_pixels,), dtype=cp.float32)
            peak_times = cp.empty((n_pixels,), dtype=cp.float32)
            channel_num = 0
            t_start = 0.0
            duration = 100.0
            camera.apply_peak_detection_to_channel(peak_amps, peak_times, channel_num, t_start, duration)
            # peak_amps and peak_times now contain the results for channel 0.

        """
        signals = self.get_channel_signals(channel, reshape=False)
        time_window = self._d_windows[self._windows_map[channel]:self._windows_map[channel+1]]
        mask = self.pixel_mask
        window_size = cp.int32(len(time_window))
        n_pixels = cp.int32(self.n_pixels)

        args = [
            peak_amplitudes,
            peaking_times,
            signals,
            time_window,
            cp.float32(t_start),
            cp.float32(extent),
            mask,
            window_size,
            n_pixels,
        ]

        block_size = self._n_threads
        peak_detection((self.n_pixels,),(block_size,),(*args,), shared_mem=8*block_size)

    def _prepare_time_windows(self):
        """Generate dynamic time windows and copy related data to device.
        """
        # Generate trigger time windows and base sampling windows
        # _prepare_time_windows has been called at the beginning of a event
        if self.triggered == False:
            n_tot = self.source[0].shape[0]
            if n_tot > 0:
                tmin = cp.min(self.source[0]).get()
                tmax = cp.max(self.source[0]).get()
            else:
                tmin = 0
                tmax = max([self.channels_time_resolution[i] for i in self.trigger_channels])
            
            self.time_windows = [None]*self.n_channels
            self._windows_map = np.empty((self.n_channels+1,), dtype=np.int32)
            windows_size = 0
            self._windows_map[0] = 0
            for i in range(self.n_channels):
                if i in self.trigger_channels:
                    t0 = tmin + self.trigger_window_start_offset[i]
                    t1 = tmax + self.trigger_window_end_offset[i]
                    self.time_windows[i] = np.arange(t0, t1, self.channels_time_resolution[i], dtype=np.float32)
                    windows_size += self.time_windows[i].shape[0]
                elif i in self.sampling_channels:
                    sampling_channel_index = self.sampling_channels.index(i)
                    sampling_window_extent = self.sampling_window_extent[sampling_channel_index]
                    self.time_windows[i] = np.arange(0, sampling_window_extent, self.channels_time_resolution[i], dtype=np.float32)
                    # If the camera trigger is disabled a trigger time must been provided
                    if self.trigger_time is not None:
                        self.time_windows[i] += t0
                    windows_size += self.time_windows[i].shape[0]
                else:
                    auxiliary_channel_index = self.auxiliary_channels.index(i)
                    auxiliary_window_extent = self.auxiliary_window_extent[auxiliary_channel_index]
                    self.time_windows[i] = np.arange(0, auxiliary_window_extent, self.channels_time_resolution[i], dtype=np.float32)
                    windows_size += self.time_windows[i].shape[0]
                
                self._windows_map[i+1] = windows_size
            
            # Allocate space on the device for time windows
            self._d_windows = cp.empty((windows_size,), dtype=cp.float32)
            # Copy mapping to device
            self._d_windows_map = cp.asarray(self._windows_map, dtype=cp.int32)
        
        # If the camera trigger is enabled the trigger time is known
        # only if the read-out has been triggered
        if self.triggered == True:
            for i in range(self.n_channels):
                if i in self.trigger_channels:
                    continue
                elif i in self.sampling_channels:
                    sampling_channel_index = self.sampling_channels.index(i)
                    t0 = self.trigger_time + self.sampling_delay[sampling_channel_index]
                else:
                    auxiliary_channel_index = self.auxiliary_channels.index(i)
                    t0 = self.trigger_time + self.auxiliary_delay[auxiliary_channel_index]
                self.time_windows[i] += t0

    # Perform all actions
    def simulate_response(self, source=None):
        """Simulates the camera response to a given source of photo-electrons.

        This method controls the simulation process:

          1. Getting the source.
          2. Preparing time windows for trigger channels.
          3. Triggering actions (pre, trigger, post).
          4. Preparing time windows for channels to be sampled.
          5. Sampling actions (pre, sampling, post).

        If :py:attr:`enable_camera_trigger` is True, the :py:meth:`pre_trigger_action`,
        :py:meth:`trigger_action`, and :py:meth:`post_trigger_action` methods are called in
        sequence.
        If a trigger occurs (as determined by :py:meth:`trigger_action`),
        or if camera triggering is disabled, the sampling phase starts.

        Parameters
        ----------
        source : Tuple[cp.ndarray, cp.ndarray], optional
            A tuple containing:

              - **pes**: A 1D CuPy array of shape (total_n_pes,) containing the
                arrival times of all photo-electrons.

              - **pes_map**: A 1D CuPy array of shape (n_pixels + 1,) representing
                a mapping to access the arrival times for each pixel.  The arrival
                times for pixel `i` are given by `pes[pes_map[i]:pes_map[i+1]]`.
            
            If None only the background is simulated.

        Raises
        ------
        RuntimeError
            If camera triggering is disabled (:py:attr:`enable_camera_trigger` is False)
            and :py:attr:`trigger_time` is not set (i.e., is None). This ensures that
            a valid trigger time is available for the sampling phase.

        Notes
        -----
        The method also increments an event counter (`self._event_counter`)
        after each simulation.

        If :py:attr:`increment_seed` is true, the random seed (:py:attr:`seed`) is
        incremented by :py:attr:`n_pixels` after each simulation. This helps
        ensure that the simulation of different events produces different random
        values (e.g. background and cross-talk generation).

        The following methods are called during the simulation:

            1. If camera triggering is enabled:

                a. :py:meth:`pre_trigger_action()`: Actions before the trigger.
                b. :py:meth:`trigger_action()`: Performs the trigger logic (should set :py:attr:`triggered` and :py:attr:`trigger_time` attributes).
                c. :py:meth:`post_trigger_action()`: Actions after the trigger.
            
            2. If :py:attr:`triggered` is True or camera triggering is disabled (and a custom
               trigger time is provided):

                a. :py:meth:`pre_sampling_action()`: Actions before sampling.
                b. :py:meth:`sampling_action()`: Performs the sampling.
                c. :py:meth:`post_sampling_action()`: Actions after sampling.
            
        """
        if not self._waveforms_are_initialized:
            raise(RuntimeError('Waveforms have not been initialized properly.'))
        
        if self.timer.active:
            cp.cuda.stream.get_current_stream().synchronize()
            t0 = time.time()
        
        self.source = source
        
        self.triggered = False
        if self.enable_camera_trigger:
            self.trigger_time = None

        if self.timer.active:
            cp.cuda.stream.get_current_stream().synchronize()
            t1 = time.time()
            self.timer.add_entry('simulate_response', 'get source', t1-t0)

        if not self.fixed_time_windows:
            self._prepare_time_windows()

        if self.timer.active:
            cp.cuda.stream.get_current_stream().synchronize()
            t2 = time.time()
            self.timer.add_entry('simulate_response', 'compute trigger windows', t2-t1)

        if self.enable_camera_trigger:
            # Where to compute trigger signals
            self.pre_trigger_action()

            if self.timer.active:
                cp.cuda.stream.get_current_stream().synchronize()
                t3 = time.time()
                self.timer.add_entry('simulate_response', 'pre_trigger_action', t3-t2)

            # Where to perform trigger and update self.triggered attribute
            self.trigger_action()

            if self.timer.active:
                cp.cuda.stream.get_current_stream().synchronize()
                t4 = time.time()
                self.timer.add_entry('simulate_response', 'trigger_action', t4-t3)
            
            self.post_trigger_action()

            if self.timer.active:
                cp.cuda.stream.get_current_stream().synchronize()
                t5 = time.time()
                self.timer.add_entry('simulate_response', 'post_trigger_action', t5-t4)
        else:
            if self.timer.active:
                self.timer.add_entry('simulate_response', 'pre_trigger_action', 0)
                self.timer.add_entry('simulate_response', 'trigger_action', 0)
                self.timer.add_entry('simulate_response', 'post_trigger_action', 0)
                t5 = t2

        if not self.enable_camera_trigger or self.triggered:
            self._triggered_events.append(self._event_counter)

            if self.trigger_time is None:
                raise(RuntimeError("A trigger time must be provided if the camera trigger is disabled."))

            if not self.fixed_time_windows:
                self._prepare_time_windows()

            if self.timer.active:
                cp.cuda.stream.get_current_stream().synchronize()
                t6 = time.time()
                self.timer.add_entry('simulate_response', 'compute sampling windows', t6-t5)

            # Where to compute sampling signals
            self.pre_sampling_action()

            if self.timer.active:
                cp.cuda.stream.get_current_stream().synchronize()
                t7 = time.time()
                self.timer.add_entry('simulate_response', 'pre_sampling_action', t7-t6)
            
            self.sampling_action()

            if self.timer.active:
                cp.cuda.stream.get_current_stream().synchronize()
                t8 = time.time()
                self.timer.add_entry('simulate_response', 'sampling_action', t8-t7)
            
            self.post_sampling_action()

            if self.timer.active:
                cp.cuda.stream.get_current_stream().synchronize()
                t9 = time.time()
                self.timer.add_entry('simulate_response', 'post_sampling_action', t9-t8)
        
        self._event_counter += 1

        if self.increment_seed:
            self.base_seed = self.base_seed + self.n_pixels

class CherenkovSiPMCamera(CherenkovCamera):
    """
    Base class for a Cherenkov camera that uses Silicon Photo-Multipliers (SiPMs).
    Inherits from CherenkovCamera and implements virtual methods specific to SiPMs.
    """
    def __init__(
            self,
            n_pixels: int,
            waveforms: Union[List[Waveform], Tuple[Waveform]],
            trigger_channels: List[int],
            sampling_channels: List[int],
            channels_time_resolution: List[float],
            pde: PhotonDetectionEfficiency = None,
            geometry: ModularSiPMCameraGeometry = None
        ):
        # Initialize the base class
        super().__init__(
            n_pixels,
            waveforms,
            trigger_channels,
            sampling_channels,
            channels_time_resolution
        )
        self.geometry = geometry
        self.pde = pde

        # SiPM-wise parameters
        self.cross_talk = cp.zeros((self.n_pixels,), dtype=cp.float32)
        self.sigma_ucells = cp.zeros((self.n_pixels,), dtype=cp.float32)
        self.background_rate = cp.zeros((self.n_pixels,), dtype=cp.float32)

        self.afterpulse = cp.zeros((self.n_pixels,), dtype=cp.float32)
        self.afterpulse_inv_tau = cp.zeros((self.n_pixels,), dtype=cp.float32)
        self.inv_tau_recovery = cp.zeros((self.n_pixels,), dtype=cp.float32)

        self.simulate_ucells = False
        self.n_ucells = cp.zeros((self.n_pixels,), dtype=cp.int32)

        self._n_threads = 128
    
    @property
    def simulate_ucells(self):
        return self._simulate_ucells

    @simulate_ucells.setter
    def simulate_ucells(self, value):
        self._simulate_ucells = value
        # Update the default empty source
        if len(self.source[0]) == 0:
            self.source = None

    @property
    def cross_talk(self):
        return self._cross_talk

    @cross_talk.setter
    def cross_talk(self, xt):
        self._cross_talk = cp.asarray(xt, dtype=cp.float32)

    @property
    def sigma_ucells(self):
        return self._sigma_ucells

    @sigma_ucells.setter
    def sigma_ucells(self, s_ucells):
        self._sigma_ucells = cp.asarray(s_ucells, dtype=cp.float32)

    @property
    def background_rate(self):
        return self._background_rate

    @background_rate.setter
    def background_rate(self, bkg):
        self._background_rate = cp.asarray(bkg, dtype=cp.float32)
        self._inv_background_rate = cp.where(bkg>1e-9, 1./bkg, 1e9)

    @property
    def pixel_mask(self):
        return self._pixel_mask

    @pixel_mask.setter
    def pixel_mask(self, mask):
        self._pixel_mask = cp.asarray(mask, dtype=cp.int32)

    def compute_signals(self, channels=None):
        """Compute SiPM signals for the specified channels.

        Parameters
        ----------
        channels : List[int], optional
            Channels IDs, by default computes all channels.
        """
        # Skip computation of channels not in `channels`
        if channels is None:
            skip_channels = cp.zeros((self.n_channels,), dtype=cp.bool_)
        else:
            skip_channels = cp.asarray([i not in channels for i in range(self.n_channels)], dtype=cp.bool_)

        if self.timer.active:
            cp.cuda.stream.get_current_stream().synchronize()
            t0 = time.time()
            self.timer.add_section(f'compute_signals{str(channels)}')

        # Fixed time windows: copy to device all windows but only at the first event
        if self.fixed_time_windows and self._event_counter == 0:
            self._d_windows, self._windows_map, self._d_windows_map = self._get_time_windows_parameters()
        
        # Dynamic time windows
        # Windows calculation is handled by `_prepare_time_windows` base class method in two steps:
        #   1) Always at the beginning of a event:
        #       - define all non-sampling windows
        #       - define the sampling windows in a range 
        #           - [0,sampling_extent] if the camera trigger is enabled;
        #           - [trigger_time,trigger_time+sampling_extent] if the camera trigger is disabled;
        #       - define the mapping (since the size of the windows is known)
        #         on the host (`self._windows_map`) and on the device (`self._d_windows_map`)
        #   2) if the camera is triggered:
        #       - add trigger_time to the sampling time window 
        #         [0,sampling_extent] -> [trigger_time,trigger_time+sampling_extent] 
        #
        # Here we copy to device only the time windows corresponding
        # to the channels that are going to be computed
        if not self.fixed_time_windows:
            for i in channels:
                self._d_windows[self._windows_map[i]:self._windows_map[i+1]] = cp.asarray(self.time_windows[i], dtype=cp.float32)
        
        self._time_windows_params = [
            self._d_windows,
            self._d_windows_map
        ]

        if self.timer.active:
            cp.cuda.stream.get_current_stream().synchronize()
            t1 = time.time()
            self.timer.add_entry(f'compute_signals{str(channels)}', 'windows_parameters', t1-t0)

        # Signals
        signals_map = self._d_windows_map*self.n_pixels
        signals_n_samples = self._windows_map[-1]*self.n_pixels

        if self.timer.active:
            cp.cuda.stream.get_current_stream().synchronize()
            t2 = time.time()
            self.timer.add_entry(f'compute_signals{str(channels)}', 'signals mapping', t2-t1)
                
        # Allocate only if the array size is different
        # (i.e.: do not re-allocate when computing signals of the same event)
        # Maybe it would be better to allocate only if a bigger array is needed
        # and access signals only through `get_channel_signals` method
        if self.signals.shape[0] != signals_n_samples:
            self.signals = cp.empty((signals_n_samples,), dtype=cp.float32)
        
        if self.timer.active:
            cp.cuda.stream.get_current_stream().synchronize()
            t3 = time.time()
            self.timer.add_entry(f'compute_signals{str(channels)}', 'signals allocation', t3-t2)

        signals_params = [
            self.signals,
            signals_map,
            skip_channels,
            cp.int32(self.n_channels)
        ]

        if self.timer.active:
            cp.cuda.stream.get_current_stream().synchronize()
            t4 = time.time()
            self.timer.add_entry(f'compute_signals{str(channels)}', 'signals_parameters', t4-t3)

        # SiPMs
        sipms_params = [
            self.cross_talk,
            self.sigma_ucells,
            self.pixel_mask,
            self.afterpulse,
            self.afterpulse_inv_tau,
            self.inv_tau_recovery
        ]

        if self.timer.active:
            cp.cuda.stream.get_current_stream().synchronize()
            t5 = time.time()
            self.timer.add_entry(f'compute_signals{str(channels)}', 'sipm_parameters', t5-t4)

        # Background
        if self.fixed_time_windows and self._event_counter == 0:
            bkg_start = min([tw[0].get() if isinstance(tw, cp.ndarray) else tw[0] for tw in self.time_windows]) - self._max_wave_extent
            bkg_end = max([tw[-1].get() if isinstance(tw, cp.ndarray) else tw[-1] for tw in self.time_windows])
            self._background_params = [
                self._inv_background_rate,
                cp.float32(bkg_start),
                cp.float32(bkg_end)
            ]

        elif not self.fixed_time_windows:
            # dynamic time windows are np.ndarrays
            t0 = min([self.time_windows[i][0] for i in self.trigger_channels])
            t1 = max([self.time_windows[i][-1] for i in self.trigger_channels])
            bkg_start = min(t0, t0+min(self.sampling_delay+self.auxiliary_delay)) - self._max_wave_extent
            bkg_end = max(t1, t0+max(self.sampling_delay+self.auxiliary_delay)+max(self.sampling_window_extent+self.auxiliary_window_extent))
            self._background_params = [
                self._inv_background_rate,
                cp.float32(bkg_start),
                cp.float32(bkg_end)
            ]

        if self.timer.active:
            cp.cuda.stream.get_current_stream().synchronize()
            t6 = time.time()
            self.timer.add_entry(f'compute_signals{str(channels)}', 'background_parameters', t6-t5)

        n_pixels = cp.int32(self.n_pixels)

        if self._event_counter == 0:
            self.__gains = cp.concatenate(self.channels_gain, dtype=cp.float32)

        if self.simulate_ucells:
           self._params = [
                *self._time_windows_params,
                *signals_params,
                self.source[0],
                self.source[1],
                self.source[2],
                self.n_ucells,
                *self._waveforms_params,
                self.__gains,
                *sipms_params,
                *self._background_params,
                self.seed,
                n_pixels
            ]   
        else:
            self._params = [
                *self._time_windows_params,
                *signals_params,
                self.source[0],
                self.source[1],
                *self._waveforms_params,
                self.__gains,
                *sipms_params,
                *self._background_params,
                self.seed,
                n_pixels
            ]   

        if self.timer.active:
            cp.cuda.stream.get_current_stream().synchronize()
            t7 = time.time()
            self.timer.add_entry(f'compute_signals{str(channels)}', 'prepare_parameters', t7-t6)

        max_window_size = max([len(self.time_windows[i]) for i in channels])
        
        if self.simulate_ucells:
            max_ucell_size = cp.max(self.n_ucells).get() // 2 + 1 # __half
            shared_mem_size = (max_window_size + self._n_threads*2 + max_ucell_size + 32 + 4)*4 
            sipm_signals_w_ucells(
                (self.n_pixels*self.n_channels,),
                (self._n_threads,),
                tuple(self._params),
                shared_mem=shared_mem_size
            )
        else:
            shared_mem_size = (max_window_size + self._n_threads*3 + 32)*4 
            sipm_signals(
                (self.n_pixels*self.n_channels,),
                (self._n_threads,),
                tuple(self._params),
                shared_mem=shared_mem_size
            )

        if self.timer.active:
            cp.cuda.stream.get_current_stream().synchronize()
            t8 = time.time()
            self.timer.add_entry(f'compute_signals{str(channels)}', 'compute_signals', t8-t7)

    def pre_trigger_action(self):
        """Compute SiPM signals for trigger channels.
        """
        self.compute_signals(self.trigger_channels)

    @abstractmethod
    def trigger_action(self):
        pass

    def pre_sampling_action(self):
        """Compute SiPM signals for sampling channels.
        """
        self.compute_signals(self.sampling_channels)

    @abstractmethod
    def sampling_action(self):
        pass
    
    def plot_modules(self, plotf, figsize=10, sep_factor=1.25, skip=None):
        """Plot a subplot for each module. See :py:func:`iactsim.visualization.plot_sipm_modules`.
        """
        from ..visualization._sipm_camera_plots import plot_sipm_modules
        return plot_sipm_modules(self, plotf, figsize, sep_factor, skip)
    
    def plot_module_pixels(self, plotf, subplot_size=4):
        """Plot a subplot for each pixel within a module. See :py:func:`iactsim.visualization.plot_sipm_module_pixels`.
        """
        from ..visualization._sipm_camera_plots import plot_sipm_module_pixels
        return plot_sipm_module_pixels(self, plotf, subplot_size)


# Copyright (C) 2024- Davide Mollica <davide.mollica@inaf.it>
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of iactsim.
#
# iactsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# iactsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iactsim.  If not, see <https://www.gnu.org/licenses/>.

from pathlib import Path
import time
from typing import Union, Dict, Any
import yaml
import random

import matplotlib.pyplot as plt
import numpy as np
import cupy as cp
from tqdm.auto import tqdm

from ...electronics._camera_geometry import ModularSiPMCameraGeometry
from ...electronics._camera import CherenkovSiPMCamera
from ...electronics._waveforms import Waveform
from ...electronics._pde import PhotonDetectionEfficiency
from ...electronics.signals.discriminator_signals._disciminator_kernels import (
    last_trigger_time_counter_from_camera_trigger,
    count_pixel_triggers
)
from ...electronics.signals.trigger_logic._trigger_kernels import (
    topological_camera_trigger,
    count_topological_camera_triggers
)
from ...electronics.signals.sampling_signals import digitize
from ...electronics.sources._poisson_flash import generate_poisson_flash

from ...io._path_loader import PathLoader

class AstriCameraGeometry(ModularSiPMCameraGeometry):
    """ASTRI camera geometry class.

    See Also
    --------
        :py:class:`iactsim.electronics.ModularSiPMCameraGeometry`: 
    """
    # Number of pdms per row
    _rows = np.array([3, 5, 7, 7, 7, 5, 3])

    # Maxium number of pdm in a row
    _max_per_row = np.max(_rows)

    #_pdms_below[i] is the number of pdms below the row i
    _pdms_below = np.append(np.array([0]), np.cumsum(_rows))

    # Total number of pdm
    _number_of_pdm = _pdms_below[-1]

    # Number of pixels per PDM side
    _n_pix_pdm = 8

    def __init__(self):
        super().__init__()

        # TODO: should it be used?
        self.position = np.array([0,0,0])
        """Position of the center of the central SiPM module (PDM 19) in the focal surface reference system."""

        self.pixels_separation = 0.2

        self.pixel_active_side = 6.975
    
    @property
    def body_dimensions(self):
        return np.asarray([0,0,0])
    
    @property
    def position(self):
        return self.__position
    
    @position.setter
    def position(self, p):
        self.__position = p
        self.__pdms_p, self.__pdms_n = self.pdms_position_and_rotation()
    
    @property
    def modules_p(self):
        return self.__pdms_p

    @property
    def modules_n(self):
        return self.__pdms_n

    @property
    def module_side(self):
        return self.pixel_active_side*self._n_pix_pdm + (self._n_pix_pdm-1)*self.pixels_separation

    @property
    def pixel_active_side(self):
        return self._pixel_active_side
    
    @pixel_active_side.setter
    def pixel_active_side(self, a_side):
        self._pixel_active_side = a_side

    @property
    def pixels_separation(self):
        return self._pixels_separation

    @pixels_separation.setter
    def pixels_separation(self, a_sep):
        self._pixels_separation = a_sep
    
    def pdmid_to_indices(self, k):
        # Out of range check
        if k<1 or k>self._number_of_pdm: 
            return None, None
        for j in range(len(self._rows)): 
            # Check if the pdm is in the row j
            if k <= self._pdms_below[j+1]:
                # Position of the pdm k starting from the first pdm of the row
                on_row_pos = k - self._pdms_below[j] - 1
                # Missing pdms at the beginning of the row
                n_missing = int((self._max_per_row - self._rows[j])/2)
                # Column i
                i = on_row_pos + n_missing
                return i, j

    def pdms_position_and_rotation(self):
        """Compute and return ASTRI photon detection modules position and normal vector with respect to the focal surface reference system."""

        pdm_angular_sep = 3.1945 # degree
        positions = np.empty((37,3), dtype=np.float64)
        normals = np.empty((37,3), dtype=np.float64)

        for k in range(37):
            i, j = self.pdmid_to_indices(k+1)
            
            beta = pdm_angular_sep * (i - 7//2)
            gamma = pdm_angular_sep * (j - 7//2)
            
            beta_rad = np.deg2rad(beta)
            gamma_rad = np.deg2rad(gamma)
            
            z = 1060 * np.cos(gamma_rad) * np.cos(beta_rad)
            x = 1060 * np.cos(gamma_rad) * np.sin(beta_rad)
            y = 1060 * np.sin(gamma_rad)

            positions[k,:] = np.asarray([x,y,z-1060]) + np.asarray(self.position)
            normals[k,:] = np.asarray([x,y,z])/1060.
        
        return positions, normals

class AstriCherenkovCamera(CherenkovSiPMCamera):
    """
    ASTRI Cherenkov camera class.

    See Also
    --------
        :py:class:`iactsim.electronics.CherenkovCamera`:
        :py:class:`iactsim.electronics.CherenkovSiPMCamera`:
    """
    _left_asic = np.asarray([0, 1, 2, 3, 8, 9, 10, 11, 16, 17, 18, 19, 24, 25, 26, 27, 32, 33, 34, 35, 40, 41, 42, 43, 48, 49, 50, 51, 56, 57, 58, 59])
    _right_asic = np.asarray([4, 5, 6, 7, 12, 13, 14, 15, 20, 21, 22, 23, 28, 29, 30, 31, 36, 37, 38, 39, 44, 45, 46, 47, 52, 53, 54, 55, 60, 61, 62, 63])

    def __init__(
            self,
        ):
        trigger_channels = [0]
        sampling_channels = [1,2]
        geometry = AstriCameraGeometry()
        waveforms = [Waveform()]*4
        channels_time_resolution = [1,1,1,1]
        pde = PhotonDetectionEfficiency()
        n_pixels = 2368

        # Initialize the base class
        super().__init__(
            n_pixels,
            waveforms,
            trigger_channels,
            sampling_channels,
            channels_time_resolution,
            pde,
            geometry,
        )        

        self.channels_gain[0] = cp.full((self.n_pixels,), 1., dtype=cp.float32)
        self.channels_gain[1] = cp.full((self.n_pixels,), 100, dtype=cp.float32)
        self.channels_gain[2] = cp.full((self.n_pixels,), 10, dtype=cp.float32)
        self.channels_gain[3] = self.channels_gain[0]

        # "Time" discriminator
        self.threshold_1 = cp.full((self.n_pixels,), 3, cp.float32)
        self.threshold_1_offset = cp.zeros((self.n_pixels,), dtype=cp.float32)
        self.n_contiguous = 5
        self.discriminator1_signals = None
        self.discriminator1_time_over_threshold = 1 # ns
        
        # Sampling info
        self.enable_peak_detection = True
        # ADCs baseline
        self.hg_adc_offset = cp.full((self.n_pixels,), 2**13, dtype=cp.float32)
        self.lg_adc_offset = cp.full((self.n_pixels,), 2**13, dtype=cp.float32)
        # ADCs noise
        self.hg_adc_noise = cp.full((self.n_pixels,), 10, dtype=cp.float32)
        self.lg_adc_noise = cp.full((self.n_pixels,), 2, dtype=cp.float32)
        self.adc_resolution = 14
        # Peak detection window extent
        self.sampling_window_extent = [45.,45.]
        # Time delay between a PDM triger and the peak-detector activation
        self.sampling_delay = [32., 32.]

        self.auxiliary_delay = [-128.]
        self.auxiliary_window_extent = [256.]

        self.peak_detection_window_extent = 32.

        self.sigma_ucells = np.full((self.n_pixels,), 0.01) 
        self.cross_talk = np.full((self.n_pixels,), 0.03)
        
        # Extend the trigger window by the trigger waveform peaking time
        self.trigger_window_end_offset = 30.

        # Calibration system
        self.blue_laser_illumination = cp.full((self.n_pixels,), 2, cp.float32)
        self.green_laser_illumination = cp.full((self.n_pixels,), 2, cp.float32)
        self.dark_rate = cp.full((self.n_pixels,), 0.001, cp.float32)

        # Seed
        self.increment_seed = True

        # Data
        self._peak_time_hg = cp.empty((self.n_pixels,), dtype=cp.float32)
        self._peak_time_lg = cp.empty((self.n_pixels,), dtype=cp.float32)
        self._peak_amplitude_hg = cp.empty((self.n_pixels,), dtype=cp.float32)
        self._peak_amplitude_lg = cp.empty((self.n_pixels,), dtype=cp.float32)
        self._trigger_time = []
        self._triggered_pdm = []
        self._pixel_trigger_time = []
        self._high_gain = []
        self._low_gain = []

        self.register_pixel_trigger_time = True

        self.n_ucells = cp.full((self.n_pixels,), 8649, dtype=cp.int32)
        self.inv_tau_recovery = cp.full((self.n_pixels,), 1./200, dtype=cp.float32)

        self.show_progress = False
        
        self.restart()

        self.threshold = 3

    def restart(self):
        super().restart()
        self._event_counter = 0
        self._high_gain = []
        self._low_gain = []
        self._trigger_time = []
        self._triggered_pdm = []
        self._pixel_trigger_time = []
        self.timer.clear()

        left = self._left_asic
        right = self._right_asic
        self.pe_to_dac = cp.empty((37,64), dtype=cp.float32)
        peq = self.channels_gain[0].reshape(37,64)
        for k in range(37):
            k_peq = peq[k]
            self.pe_to_dac[k,left] = cp.nanmedian(k_peq[left])
            self.pe_to_dac[k,right] = cp.nanmedian(k_peq[right])

    @property
    def peak_detection_window_extent(self):
        return self._peak_detection_window_extent
    
    @peak_detection_window_extent.setter
    def peak_detection_window_extent(self, a_extent):
        # clock maxium delay
        dt = 1e3/200. + 1e3/300.
        self.sampling_window_extent[0] = a_extent + dt
        self.sampling_window_extent[1] = a_extent + dt
        self._peak_detection_window_extent = a_extent

    @property
    def threshold(self):
        return self._threshold

    @threshold.setter
    def threshold(self, pe):
        """Compute the disciminator thresholds for a given global threshold.
        That is, performs the so-called pe-to-DAC conversion.

        Parameters
        ----------
        pe : float
            Global threshold in photo-electrons.
        """
        self.threshold_1 = cp.round(pe*self.pe_to_dac)
        self._threshold = pe

    @property
    def triggered_pdm(self):
        if len(self._triggered_pdm) == 0:
            return None
        else:
            return np.asarray(self._triggered_pdm)

    @property
    def high_gain(self):
        if len(self._high_gain) == 0:
            return None
        else:
            return np.stack([hg.get() for hg in self._high_gain])

    @property
    def low_gain(self):
        if len(self._low_gain) == 0:
            return None
        else:
            return np.stack([lg.get() for lg in self._low_gain])

    @property
    def pixel_trigger_time(self):
        if len(self._pixel_trigger_time) == 0:
            return None
        else:
            return np.stack([tt.get() for tt in self._pixel_trigger_time])

    @property
    def adc_resolution(self):
        return self._adc_resolution

    @adc_resolution.setter
    def adc_resolution(self, a_res):
        self._adc_max_value = cp.int32(2**a_res-1)
        self._adc_resolution = a_res
    
    @property
    def threshold_1(self):
        return self._threshold_1

    @threshold_1.setter
    def threshold_1(self, thrs):
        self._threshold_1 = cp.asarray(thrs, dtype=cp.float32)

    @property
    def hg_adc_noise(self):
        return self._hg_adc_noise

    @hg_adc_noise.setter
    def hg_adc_noise(self, noise):
        self._hg_adc_noise = cp.asarray(noise, dtype=cp.float32)

    @property
    def hg_adc_offset(self):
        return self._hg_adc_offset

    @hg_adc_offset.setter
    def hg_adc_offset(self, offset):
        self._hg_adc_offset = cp.asarray(offset, dtype=cp.float32)

    @property
    def lg_adc_noise(self):
        return self._lg_adc_noise

    @lg_adc_noise.setter
    def lg_adc_noise(self, noise):
        self._lg_adc_noise = cp.asarray(noise, dtype=cp.float32)

    @property
    def lg_adc_offset(self):
        return self._lg_adc_offset

    @lg_adc_offset.setter
    def lg_adc_offset(self, offset):
        self._lg_adc_offset = cp.asarray(offset, dtype=cp.float32)
        
    def trigger_action(self):
        """Generate a camera trigger.
        """
        if self.timer.active:
            cp.cuda.stream.get_current_stream().synchronize()
            t0 = time.time()
            self.timer.add_section('trigger_action')
            
        self.discriminator1_signals = self.apply_ideal_discriminator_to_channel(
            0,
            self.threshold_1,
            self.threshold_1_offset,
            self.discriminator1_time_over_threshold/self.channels_time_resolution[0]
        )

        if self.timer.active:
            cp.cuda.stream.get_current_stream().synchronize()
            t1 = time.time()
            self.timer.add_entry('trigger_action', 'discriminator', t1-t0)
        
        n_pdms = len(self.geometry.modules_p)
        
        pdm_dimension = cp.int32(self.geometry.module_side/ self.geometry.pixel_active_side)
        window_size = cp.int32(len(self.time_windows[0]))
        n_contiguous = cp.int32(self.n_contiguous)
        self.pdms_trigger_signals = cp.empty((n_pdms,len(self.time_windows[0])), dtype=cp.float32)
        trigger = cp.full((2,), -1, dtype=cp.int32)

        # Better to allocate this once (change this only if n_pixels, n_pdms or _n_threads change)
        ints_per_thread = self.n_pixels + (self.n_pixels // 32) + 2
        total_elements = n_pdms * self._n_threads * ints_per_thread
        global_buffer = cp.empty(total_elements, dtype=cp.int32)

        if self.timer.active:
            cp.cuda.stream.get_current_stream().synchronize()
            t2 = time.time()
            self.timer.add_entry('trigger_action', 'copy to device', t2-t1)

        trigger_args = (
            self.discriminator1_signals,
            n_contiguous,
            self.pdms_trigger_signals,
            trigger,
            global_buffer,
            window_size,
            pdm_dimension,
            cp.int32(n_pdms),
            self.seed[0]
        )
        topological_camera_trigger((n_pdms,), (128,), trigger_args)

        if self.timer.active:
            cp.cuda.stream.get_current_stream().synchronize()
            t3 = time.time()
            self.timer.add_entry('trigger_action', 'camera trigger', t3-t2)
        
        t_index, pdm_id = trigger.get()
        if t_index >= 0:
            self.triggered = True
            self.trigger_time = self.time_windows[0][t_index]
            self._triggered_pdm.append(pdm_id+1) # count from 1 to match the usual ID
            self._trigger_time.append(self.trigger_time)
        else:
            self.trigger_time = None

        if self.timer.active:
            cp.cuda.stream.get_current_stream().synchronize()
            t4 = time.time()
            self.timer.add_entry('trigger_action', 'copy to host', t4-t3)
    
    def sampling_action(self):
        """Perform peak detection and pixels trigger time measurement.
        """
        dt = 0.;#np.random.uniform(0, 1e3/200 + 1e3/300)
        self._t0_peak_detection_hg = self.trigger_time + self.sampling_delay[0] + dt
        self._t0_peak_detection_lg = self.trigger_time + self.sampling_delay[1] + dt

        if self.enable_peak_detection:
            extent_hg = self.peak_detection_window_extent
            extent_lg = self.peak_detection_window_extent
            self.apply_peak_detection_to_channel(self._peak_amplitude_hg, self._peak_time_hg, 1, self._t0_peak_detection_hg, extent_hg)
            self.apply_peak_detection_to_channel(self._peak_amplitude_lg, self._peak_time_lg, 2, self._t0_peak_detection_lg, extent_lg)
        else:
            t_slice_hg = np.abs(self.time_windows[1] - self._t0_peak_detection_hg).argmin()
            t_slice_lg = np.abs(self.time_windows[2] - self._t0_peak_detection_lg).argmin()
            self._peak_amplitude_hg = self.get_channel_signals(1,reshape=True)[:,t_slice_hg].flatten()
            self._peak_amplitude_lg = self.get_channel_signals(2,reshape=True)[:,t_slice_lg].flatten()
        
        dig_hg = cp.empty((self.n_pixels,), dtype=cp.uint16)
        hg_args = (
            dig_hg,
            self._peak_amplitude_hg,
            self.hg_adc_offset,
            self.hg_adc_noise,
            self.seed,
            self._adc_max_value,
            cp.int32(self.n_pixels),
            cp.int32(1)
        )
        dig_lg = cp.empty((self.n_pixels,), dtype=cp.uint16)
        lg_args = (
            dig_lg,
            self._peak_amplitude_lg,
            self.lg_adc_offset,
            self.lg_adc_noise,
            self.seed,
            self._adc_max_value,
            cp.int32(self.n_pixels),
            cp.int32(1)
        )
        
        digitize((self.n_pixels,),(32,),hg_args)
        digitize((self.n_pixels,),(32,),lg_args)

        self._high_gain.append(dig_hg)
        self._low_gain.append(dig_lg)

    def post_sampling_action(self):
        ## Trigger time
        if self.register_pixel_trigger_time:
            tt = cp.empty((self.n_pixels,), dtype=cp.int32)
            self.compute_signals([3])
            self._time_discriminator_signals = self.apply_ideal_discriminator_to_channel(
                3,
                self.threshold_1,
                self.threshold_1_offset,
                self.discriminator1_time_over_threshold/self.channels_time_resolution[3]
            )
            delay = 128. + self.auxiliary_delay[0]
            cam_trigger_clock = cp.int32((np.abs(self.time_windows[3] - (self.trigger_time+delay))).argmin())
            stop_clock = cp.int32(128)
            register_n_bits = cp.int32(8)
            wsize = len(self.time_windows[3])
            n_pixels = cp.int32(self.n_pixels)
            n_threads = self._n_threads
            last_trigger_time_counter_from_camera_trigger(
                (self.n_pixels,),
                (n_threads,),
                (
                    tt,
                    self._time_discriminator_signals,
                    cam_trigger_clock,
                    stop_clock,
                    register_n_bits,
                    cp.int32(wsize), 
                    n_pixels
                )
            )
            self._pixel_trigger_time.append(tt)

    def acquire_pedestal(self, n_frame):
        """Acquire pulse-height distribution with an external trigger and the current background.

        Parameters
        ----------
        n_frame : int
            number of frame to be acquired.
        """
        self.restart()

        # Save current configuration
        old_enable_camera_trigger = self.enable_camera_trigger
        old_fixed_windows = self.fixed_time_windows

        # Pedestal configuration
        self.trigger_time = 0
        self.enable_camera_trigger = False
        self.fixed_time_windows = True
        self.time_windows = [
            np.asarray([0],dtype=np.float32),
            np.arange(self.sampling_delay[0], self.sampling_delay[0]+self.sampling_window_extent[0], self.channels_time_resolution[1], dtype=np.float32),
            np.arange(self.sampling_delay[1], self.sampling_delay[1]+self.sampling_window_extent[1], self.channels_time_resolution[2], dtype=np.float32),
            np.arange(-128, 128., self.channels_time_resolution[3], dtype=np.float32),
        ]
        
        # Pedestal
        for i in tqdm(range(n_frame), disable=not self.show_progress):
            self.simulate_response(source=None)

        # Restore old configuration (not windows)
        self.enable_camera_trigger = old_enable_camera_trigger
        self.fixed_time_windows = old_fixed_windows
    
    def acquire_variance(self, n_sample_hg: int, n_sample_lg: int = None, source: tuple = None, background: cp.ndarray = None) -> tuple[cp.ndarray, cp.ndarray]:
        """Acquire variance data.

        Parameters
        ----------
        n_sample_hg : int
            Number of HG sample to be acquired.
        n_sample_lg : int, optional
            Number of LG sample to be acquired. If None use HG number of sample.
        source : int, optional
            Source to be added to the background. If it is not None, the variance is measured
            at tmin+0.9*(tmax-tmin), where tmin and tmax are the minimum and maximum arrival time of the source.
            The source arrival times should be uniformly distributed between tmin and tmax.
            At each variance sample the source arrival times are randomized again, but the number
            of discharge per pixel/microcell is the same.
        background : int, optional
            If not None, temporarly replace the background attribute. The old attribute is restored.

        Returns
        -------
        tuple(cp.ndarray, cp.ndarray)
            High gain and low gain variance arrays.

        Notes
        -----
        In order to simulate microcells with non-diffused background (e.g. starts), a naive way
        is to use the output `source` of the ray-tracing (get_camera_input=True) and randomize the arrival time
        at each variance sample. This is the default behaviour if you pass `source` to this method.
        But remember to generate the source photons with uniform (and large-spreaded) arrival times:

            ..code-block:: python

                source = sources.Source(astri_tel)
                source.arrival_times.type = 'uniform'
                source.arrival_times.tmin = 0
                source.arrival_times.tmax = 2048

        Although the timing is handle properly, the microcell IDs are not. A fix for this is planned in the future.

        """
        self.restart()

        # Save current configuration
        old_fixed_windows = self.fixed_time_windows

        if n_sample_lg is None:
            n_sample_lg = n_sample_hg

        n_sample = max(n_sample_lg, n_sample_hg)

        hg_values = cp.empty((n_sample_hg,self.n_pixels),dtype=cp.uint16)
        lg_values = cp.empty((n_sample_lg,self.n_pixels),dtype=cp.uint16)

        stream_hg = cp.cuda.Stream(non_blocking=True)
        stream_lg = cp.cuda.Stream(non_blocking=True)

        if background is not None:
            old_background = self.background_rate.copy()
            self.background_rate = background
        
        randomize_times = False
        tmin = 0
        tmax = 0
        if source is not None:
            randomize_times = True
            tmin = source[0].min().get()
            tmax = source[0].max().get()

        # Pedestal configuration
        
        self.fixed_time_windows = True
        t0 = tmin + 0.9*(tmax-tmin)
        self.time_windows = [
            np.asarray([t0],dtype=np.float32),
            np.asarray([t0],dtype=np.float32),
            np.asarray([t0],dtype=np.float32),
            np.asarray([t0],dtype=np.float32),
        ]
        
        self.source = None
        for i in tqdm(range(n_sample), disable=not self.show_progress):
            if randomize_times:
                self.source = (
                    cp.random.uniform(tmin,tmax,size=source[0].shape,dtype=cp.float32),
                    *list(source)[1:]
                )
            
            channels = []
            if i < n_sample_hg:
                channels.append(1)
            if i < n_sample_lg:
                channels.append(2)
            
            self.compute_signals(channels)

            if i < n_sample_hg:
                hg_args = (
                    hg_values[i,:],
                    self.get_channel_signals(1),
                    self.hg_adc_offset,
                    self.hg_adc_noise,
                    self.seed,
                    self._adc_max_value,
                    cp.int32(self.n_pixels),
                    cp.int32(1)
                )
                with stream_hg:
                    digitize((self.n_pixels,),(32,),hg_args)
            
            if i < n_sample_lg:
                lg_args = (
                    lg_values[i,:],
                    self.get_channel_signals(2),
                    self.lg_adc_offset,
                    self.lg_adc_noise,
                    self.seed,
                    self._adc_max_value,
                    cp.int32(self.n_pixels),
                    cp.int32(1)
                )

                with stream_lg:
                    digitize((self.n_pixels,),(32,),lg_args)

            self.seed += self.n_pixels
        
        self.source = None
        
        stream_hg.synchronize()
        stream_lg.synchronize()

        # Restore old configuration (not windows)
        self.fixed_time_windows = old_fixed_windows
        if background is not None:
            self.background_rate = old_background
        return cp.var(hg_values,axis=0), cp.var(lg_values,axis=0)
    
    def acquire_dark_pedestal(self, n_frame):
        """Acquire pulse-height distribution with an external trigger and the closed lids background.

        Parameters
        ----------
        n_frame : int
            number of frame to be acquired.
        """
        old_background = self.background_rate.copy()
        self.background_rate = self.dark_rate

        self.acquire_pedestal(n_frame)

        self.background_rate = old_background

    def acquire_phd(self, n_frame, background=None, laser_width=10.):
        """Acquire pulse-height distribution with pulsed flash light.

        Parameters
        ----------
        n_frame : int
            number of frame to be acquired.
        laser_width: int
            Duration of the laser pulse.
        """
        self.restart()
        
        # Save current configuration
        old_enable_camera_trigger = self.enable_camera_trigger
        old_fixed_windows = self.fixed_time_windows
        old_background = self.background_rate.copy()

        # PHD configuration
        if background is not None:
            self.background_rate = background
        else:
            self.background_rate = self.dark_rate
        self.enable_camera_trigger = True
        self.fixed_time_windows = True
        self.time_windows = [
            np.arange(0,  64, self.channels_time_resolution[0], np.float32),
            np.arange(0, 256, self.channels_time_resolution[1], np.float32),
            np.arange(0, 256, self.channels_time_resolution[2], np.float32),
            np.arange(0, 256, self.channels_time_resolution[3], np.float32),
        ]
        for i in tqdm(range(n_frame), disable=not self.show_progress):
            ucells = None
            if self.simulate_ucells:
                ucells = self.n_ucells
            source = generate_poisson_flash(self.n_pixels, self.blue_laser_illumination, 0., laser_width, n_ucells=ucells, seed=None)
            self.simulate_response(source)

        # Restore previous configuration (but not time windows)
        self.enable_camera_trigger = old_enable_camera_trigger
        self.fixed_time_windows = old_fixed_windows
        self.background_rate = old_background
    
    def threshold_scan(self, start, stop, step, window_extent=1024., max_duration=1e-3, maximum_count=None, background=None):
        """Measure the camera trigger rate as a function of the trigger threshold.

        Parameters
        ----------
        start : float
            Start threshold value.
        stop : float
            Stop threshold value.
        step : float
            Threshold step.
        window_extent : float, optional
            Extent in ns of the time window where to count triggers, by default 1024 ns.
        max_duration : float, optional
            Integration time (in seconds) for each threshold, by default 1e-3 s.
        maximum_count : int, optional
            Stop when at least ``maximum_count`` triggers have been found at a each threshold, useful to speed-up the procedure at lower thresholds. By default None.

        Returns
        -------
        (numpy.ndarray,numpy.ndarray,numpy.ndarray,numpy.ndarray)
            Thresholds and corresponding rate, rate error and integration time. Integration times can be different for each threshold if a ``maximum_count`` is provided.
        """
        # Save current configuration
        old_fixed_windows = self.fixed_time_windows
        old_threshold = self.threshold
        
        self.restart()
        self.fixed_time_windows = True
        self.time_windows = [
            np.arange(0, window_extent, 10./3, np.float32),
            np.arange(0, 32, 1, np.float32),
            np.arange(0, 32, 1, np.float32),
            np.arange(0, 32, 1, np.float32),
        ]
        window_size = self.time_windows[0].shape[0]

        # Allow signals to be computed with no source
        self.source = None

        update_background = False
        if background is not None:
            if background.ndim == 1:
                self.background_rate = cp.asarray(background)
            elif background.ndim == 2 and len(background) == (stop-start)//step+1:
                update_background = True
                background = cp.asarray(background)
            else:
                raise ValueError("`background` must be a 1D or 2D array. If 2D the shape must be (n_thresholds, n_pixels).")

        max_iterations = int(max_duration/window_extent*1e9)
        thresholds = np.arange(start, stop+step, step, dtype=np.float32)
        trigger_rate = cp.zeros(thresholds.shape, dtype=cp.int32)
        integration_time = np.zeros(thresholds.shape, dtype=np.float32)
        n_threads = self._n_threads

        if self.timer.active:
            cp.cuda.stream.get_current_stream().synchronize()
            t_old = time.time()
            self.timer.add_section('trigger_scan')

        for i in tqdm(range(max_iterations), disable=not self.show_progress):
            self.compute_signals([0])
            for j in range(len(thresholds)):
                if self.timer.active:
                    cp.cuda.stream.get_current_stream().synchronize()
                    t_th_0 = time.time()

                # Skip if there are enough count at this threshold
                if maximum_count is not None:
                    if trigger_rate[j] > maximum_count:
                        continue

                if self.timer.active:
                    cp.cuda.stream.get_current_stream().synchronize()
                    t_th_1 = time.time()
                    self.timer.add_entry('trigger_scan', f'thr {thresholds[j]}: check rate', t_th_1 - t_th_0)

                # Set the threshold
                self.threshold = thresholds[j]
                if update_background:
                    self.background_rate = background[j]

                if self.timer.active:
                    cp.cuda.stream.get_current_stream().synchronize()
                    t_th_2 = time.time()
                    self.timer.add_entry('trigger_scan', f'thr {thresholds[j]}: set threshold and nsb', t_th_2 - t_th_1)

                # Get triggers
                self.trigger_action()

                if self.timer.active:
                    cp.cuda.stream.get_current_stream().synchronize()
                    t_th_3 = time.time()
                    self.timer.add_entry('trigger_scan', f'thr {thresholds[j]}: trigger action', t_th_3 - t_th_2)
                
                # Initialize count
                cam_trigger_counts = cp.zeros((1,), dtype=cp.int32)
                args = [
                    cam_trigger_counts,
                    self.pdms_trigger_signals,
                    cp.int32(window_size),
                    cp.int32(37)
                ]
                # Count triggers
                count_topological_camera_triggers((1,), (n_threads,), args, shared_mem=4*(window_size+n_threads))

                if self.timer.active:
                    cp.cuda.stream.get_current_stream().synchronize()
                    t_th_4 = time.time()
                    self.timer.add_entry('trigger_scan', f'thr {thresholds[j]}: count triggers', t_th_4 - t_th_3)

                integration_time[j] += window_extent / 1e9
                trigger_rate[j] += cam_trigger_counts[0]

                if self.timer.active:
                    cp.cuda.stream.get_current_stream().synchronize()
                    t_th_5 = time.time()
                    self.timer.add_entry('trigger_scan', f'thr {thresholds[j]}: update counts', t_th_5 - t_th_4)

                if self.timer.active:
                    cp.cuda.stream.get_current_stream().synchronize()
                    t_new = time.time()
                    self.timer.add_entry('trigger_scan', f'thr {thresholds[j]}', t_new-t_old)
                    t_old = t_new

            self.seed = self.seed + self.n_pixels
            self._event_counter += 1
        
        # Restore previous configuration (but not time windows)
        self.fixed_time_windows = old_fixed_windows
        self.threshold = old_threshold

        # Compute rate and erros
        integration_time = np.asarray(integration_time)
        trigger_rate = trigger_rate.get()
        error = np.sqrt(trigger_rate) / integration_time
        trigger_rate = trigger_rate / integration_time
        return thresholds, trigger_rate, error, integration_time

    def acquire_stairs(self, start, stop, step, integration_time=0.001, window_extent=1024.):
        """Acquire staircase curves (pixel dark trigger rate as a function of threshold).

        Parameters
        ----------
        start : float
            Start threshold value.
        stop : float
            Stop threshold value.
        step : float
            Threshold step.
        integration_time : float, optional
            Integration time (in seconds), by default 0.001 s
        window_extent : float, optional
            Extent in ns of the time window where to count triggers, by default 1024 ns.

        Returns
        -------
        (numpy.ndarray, numpy.ndarray)
            Thresholds and corresponding trigger rate per pixel.
        """
        old_background = self.background_rate.copy()
        old_fixed_windows = self.fixed_time_windows

        self.background_rate = self.dark_rate

        self.timer.clear()
        self.restart()
        self.fixed_time_windows = True
        self.time_windows = [
            np.arange(0, window_extent, 10./3., np.float32),
            np.arange(0, 32, 1, np.float32),
            np.arange(0, 32, 1, np.float32),
        ]
        window_size = self.time_windows[0].shape[0]
        
        # Allow signals to be computed with no source
        self.source = None
        
        thrs = np.arange(start, stop+step, step)
        tr_counts = cp.zeros((self.n_pixels,len(thrs)), dtype=np.int32)
        n_iterations = int(integration_time*1e9/window_extent)
        for _ in tqdm(range(n_iterations), disable=not self.show_progress):
            self.compute_signals([0])
            for i,thr in enumerate(thrs):
                counts = cp.zeros((self.n_pixels,), dtype=cp.int32)
                threshold = cp.full((self.n_pixels,), thr, dtype=cp.float32)
                disc_signals = self.apply_ideal_discriminator_to_channel(
                    0,
                    threshold,
                    self.threshold_1_offset,
                    self.discriminator1_time_over_threshold/self.channels_time_resolution[0]
                )
                count_pixel_triggers(
                    (self.n_pixels,),
                    (64,),
                    (
                        counts,
                        disc_signals,
                        cp.int32(window_size),
                        cp.int32(self.n_pixels)
                    )
                )
                tr_counts[:,i] += counts
            self.seed += self.n_pixels
            self._event_counter += 1
        
        self.background_rate = old_background
        self.fixed_time_windows = old_fixed_windows

        return thrs, tr_counts.get()

    def measure_pixel_trigger_rate(self, integration_time=0.01, window_extent=1024.):
        old_fixed_windows = self.fixed_time_windows

        self.timer.clear()
        self.restart()

        self.fixed_time_windows = True
        self.time_windows = [
            np.arange(0, window_extent, 10./3., np.float32),
            np.arange(0, 32, 1, np.float32),
            np.arange(0, 32, 1, np.float32),
            np.arange(0, 32, 1, np.float32),
        ]
        window_size = self.time_windows[0].shape[0]
        tr_counts = cp.zeros((self.n_pixels,), dtype=np.int32)
        n_iterations = int(integration_time*1e9/window_extent)
        for _ in tqdm(range(n_iterations), disable=not self.show_progress):
            self.compute_signals([0])
            counts = cp.zeros((self.n_pixels,), dtype=cp.int32)
            self.discriminator1_signals = self.apply_ideal_discriminator_to_channel(
                0,
                self.threshold_1,
                self.threshold_1_offset,
                self.discriminator1_time_over_threshold/self.channels_time_resolution[0]
            )
            count_pixel_triggers((self.n_pixels,),(128,),(counts,self.discriminator1_signals,cp.int32(window_size),cp.int32(self.n_pixels)))
            tr_counts[...] += counts
            self.seed = self.seed + self.n_pixels
            self._event_counter += 1
        
        self.fixed_time_windows = old_fixed_windows

        return tr_counts.get()/integration_time

    def configure(self, config: Union[str, Path, Dict[str, Any]]) -> None:
        """Configure the camera with a yaml configuration file or a dictionary.

        Parameters
        ----------
        config : str, path-like, or dict
            Path to the configuration file (str or pathlib.Path),
            or a dictionary containing the configuration.
        """

        if isinstance(config, (str, Path)):
            with open(config, 'r') as file:
                cam_conf = yaml.load(file, Loader=PathLoader)
        elif isinstance(config, dict):
            cam_conf = config
        else:
            raise TypeError(
                "config must be a string (path), pathlib.Path, or a dictionary, "
                f"not {type(config)}"
            )
        
        if 'waveforms' in cam_conf and cam_conf['waveforms'] is not None:
            waveforms = []
            for file in cam_conf['waveforms']:
                x, y = np.loadtxt(file).T
                wave = Waveform(x, y, 'AC')
                wave.normalize()
                waveforms.append(wave)
            self.waveforms = waveforms

        if 'pde' in cam_conf and cam_conf['pde'] is not None:
            data = np.loadtxt(cam_conf['pde']).T
            wl = data[0,:]
            pde = data[1,:]
            self.pde.value = pde
            self.pde.wavelength = wl

        if 'channels_gain' in cam_conf:
            for i,gain in enumerate(cam_conf['channels_gain']):
                if gain is not None:
                    if isinstance(gain, (int,float)):
                        array = cp.full((self.n_pixels,), gain, dtype=cp.float32)
                    else:
                        array = cp.loadtxt(gain, dtype=cp.float32)
                    if array.size != self.n_pixels:
                        raise(ValueError(f"Channel {i} gain: wrong number of values provided ({array.size}). Provide a single value or a value for each pixel ({self.n_pixels})."))
                    self.channels_gain[i] = array.flatten()

        if 'timer' in cam_conf and cam_conf['timer'] is not None:
            self.timer.active = cam_conf['timer']

        if 'seed' in cam_conf and cam_conf['seed'] is not None:
            seed = cam_conf['seed']
        else:
            seed = random.getrandbits(62)
        
        self.seed = cp.arange(0, self.n_pixels, 1, dtype=cp.uint64) + seed

        if 'simulate_ucells' in cam_conf and cam_conf['simulate_ucells'] is not None:
            self.simulate_ucells = cam_conf['simulate_ucells']
        
        assign_directly = [
            'trigger_channels',
            'trigger_window_end_offset',
            'trigger_window_start_offset',
            'sampling_channels',
            'sampling_delay',
            'sampling_window_extent',
            'peak_detection_window_extent',
            'channels_time_resolution',
            'auxiliary_delay',
            'auxiliary_window_extent',
            'adc_resolution',
            'discriminator1_time_over_threshold',
            'register_pixel_trigger_time',
            'increment_seed',
            'show_progress',
            # 'threshold' # Must be assigned after after channels_gain[0] and threshold1_offset
        ]

        check_is_value = [
            'hg_adc_noise',
            'hg_adc_offset',
            'lg_adc_noise',
            'lg_adc_offset',
            'cross_talk',
            'sigma_ucells',
            'pixel_mask',
            'background_rate',
            'dark_rate',
            'blue_laser_illumination',
            'green_laser_illumination',
            'threshold_1_offset',
            'afterpulse',
            'afterpulse_tau',
            'recovery_time',
        ]


        # self.afterpulse = cp.zeros((self.n_pixels,), dtype=cp.float32)
        # self.afterpulse_inv_tau = cp.zeros((self.n_pixels,), dtype=cp.float32)
        # self.inv_tau_recovery = cp.zeros((self.n_pixels,), dtype=cp.float32)

        for name in cam_conf.keys():

            value = cam_conf[name]
            
            attribute_name = name
            
            if value is not None:
                if name in check_is_value:
                    # Default dtype
                    dtype = cp.float32

                    if name == 'pixel_mask':
                        dtype = cp.int32
                    
                    if name == 'afterpulse_tau':
                        attribute_name = 'afterpulse_inv_tau'
                        value = 1./value
                    
                    if name == 'recovery_time':
                        attribute_name = 'inv_tau_recovery'
                        value = 1./value
                    
                    if isinstance(value, (int,float)):
                        array = cp.full((self.n_pixels,), value, dtype=dtype)
                    else:
                        array = cp.loadtxt(value, dtype=dtype)
                    
                    if array.size != self.n_pixels:
                        raise(ValueError(f"{name}: wrong number of values provided ({array.size}). Provide a single value or a value for each pixel ({self.n_pixels})."))
                    
                    setattr(self, attribute_name, array.flatten())

                if name in assign_directly:
                    setattr(self, name, value)

        # Set trigger channels info then set the threshold
        self.restart()

        if 'threshold' in cam_conf and cam_conf['threshold'] is not None:
            self.threshold = cam_conf['threshold']
        
    
    def plot_pixel_trigger_channel(self, pdm, pixel, ax=None):
        if ax is None:
            ax = plt.gca()
        pix = pixel - 1
        pdm = pdm - 1
        window = self.time_windows[0]
        gain = self.channels_gain[0][pix].get()
        signal = self.get_channel_signals(0, True)[64*pdm:64*(pdm+1)][pix].get()/gain
        disc_signal = self.discriminator1_signals.reshape(37,64,-1)[pdm,pix].get()*self.threshold_1[pdm,pix].get()/gain
        
        if self.register_pixel_trigger_time:
            window = self.time_windows[3]
            gain = self.channels_gain[3][pix].get()
            signal = self.get_channel_signals(3, True)[64*pdm:64*(pdm+1)][pix].get()/gain
            disc_signal = self._time_discriminator_signals.reshape(37,64,-1)[pdm,pix].get()*self.threshold_1[pdm,pix].get()/gain
        
        ax.plot(window, signal, ls='-', lw=2, label='Fast signal')
        ax.plot(window, disc_signal, ls='-', lw=1, color='black', label='Pixel Trigger')

        if self.triggered or not self.enable_camera_trigger:
            ax.vlines(self.trigger_time, signal.min(), signal.max(), color='black', ls='--', label='Camera Trigger')

        ax.hlines(self.threshold_1[pdm,pix].get()/gain, window[0], window[-1], ls='--', color='grey', label='Trigger threshold')

        ax.set_ylim(signal.min(), signal.max())
        
        ax.legend(title=f'PDM {pdm+1} Pixel {pix+1}')
        ax.grid(which='both')
    
    def plot_pdm_pixels_trigger_channel(self, pdm):
        from ...visualization._sipm_camera_plots import plot_sipm_module_pixels

        def plotf(pix):
            self.plot_pixel_trigger_channel(pdm, pix+1)
        
        plot_sipm_module_pixels(self, plotf, 5)

    def plot_pixel_sampling_channels(self, pdm, pixel, ax=None):
        if ax is None:
            ax = plt.gca()
        
        if self.enable_camera_trigger and not self.triggered:
            return

        pix = pixel - 1
        pdm = pdm - 1
        hg_window = self.time_windows[1]
        lg_window = self.time_windows[2]
        hg_signal = self.get_channel_signals(1, True)[64*pdm:64*(pdm+1)][pix].get()/self.channels_gain[1][pix].get()
        lg_signal = self.get_channel_signals(2, True)[64*pdm:64*(pdm+1)][pix].get()/self.channels_gain[2][pix].get()
        
        
        hg_plot = ax.plot(hg_window, hg_signal, label='HG signal')
        lg_plot = ax.plot(lg_window, lg_signal, label='LG signal')

        ymin = min(hg_signal.min(),lg_signal.min())
        ymax = max(hg_signal.max(),lg_signal.max())

        ax.set_ylim(ymin,ymax)
    
        if self.enable_peak_detection:
            t0 = self._t0_peak_detection_hg
            t1 = t0 + self.peak_detection_window_extent
            ax.fill_between(x=[t0,t1], y1=[ymin,ymin], y2=[ymax,ymax], color='grey', interpolate=True, alpha=0.2, label='PD interval')
            hg = self._peak_amplitude_hg[pdm*64+pix].get()/self.channels_gain[1][pix].get()
            lg = self._peak_amplitude_lg[pdm*64+pix].get()/self.channels_gain[2][pix].get()
            ax.hlines(hg, hg_window[0], hg_window[-1], color=hg_plot[0].get_color(), ls='--')
            ax.hlines(lg, lg_window[0], lg_window[-1], color=lg_plot[0].get_color(), ls='--')
            
        ax.legend(title=f'PDM {pdm+1} Pixel {pix+1}', framealpha=0.5)
        ax.grid(which='both')
    
    def plot_pdm_pixels_sampling_channels(self, pdm):
        from ...visualization._sipm_camera_plots import plot_sipm_module_pixels

        def plotf(pix):
            self.plot_pixel_sampling_channels(pdm, pix+1)
        
        plot_sipm_module_pixels(self, plotf, 5)


import pytest
import cupy as cp
import iactsim
import numpy as np

@pytest.fixture(scope="module")
def telescope_setup():
    """Initializes the telescope and GPU context."""
    # Spherical mirror
    mirror_curvature_radius = 20000
    plate_scale = mirror_curvature_radius/57.296/2.
    mirror = iactsim.optics.SphericalSurface(
        half_aperture=10000., 
        curvature=1./mirror_curvature_radius,
        position=(0,0,0),
        surface_type=iactsim.optics.SurfaceType.REFLECTIVE_FRONT,
        name = 'Mirror'
    )

    # Flat focal surface (5deg hexagon)
    focal_plane = iactsim.optics.FlatSurface(
        half_aperture = 5*plate_scale, 
        position = (0,0,0.5*mirror_curvature_radius),
        aperture_shape = iactsim.optics.ApertureShape.HEXAGONAL,
        surface_type=iactsim.optics.SurfaceType.SENSITIVE_BACK,
        name = 'Focal Plane'
    )

    # List of segments
    segments = []

    # Segment ID
    k = 0

    # Segments on a nXn grid, 80 total
    n = 20

    segment_distance = 2*mirror.half_aperture / (n+3)

    for i in range(n+3):
        for j in range(n+3):
            offset = [-mirror.half_aperture+segment_distance*i, -mirror.half_aperture+segment_distance*j]
            r_segment = np.sqrt(offset[0]**2+offset[1]**2)
            
            # Do not create segments outside the original mirror aperture
            if r_segment > mirror.half_aperture-segment_distance*np.sqrt(2):
                continue

            # Ideal segment position
            segment_position = [
                offset[0],
                offset[1],
                mirror.sagitta(r_segment),
            ]
            
            # Create the surface
            segment = iactsim.optics.SphericalSurface(
                curvature=mirror.curvature,
                half_aperture=0.45*segment_distance, 
                position=segment_position,
                surface_type=mirror.type,
                name = f'Segment-{k}',
                aperture_shape=iactsim.optics.ApertureShape.SQUARE,
                tilt_angles=np.random.normal(0,0.05,3),
                scattering_dispersion=0.01
            )
            
            segment.offset = offset
            
            segments.append(segment)
            k += 1
    
    segmented_os = iactsim.optics.OpticalSystem(surfaces=[focal_plane, *segments], name='SEGMENTED-TEST-OS')
    pos = (0,0,0)
    poi = (0.,0.)
    telescope = iactsim.IACT(segmented_os, position=pos, pointing=poi)
    telescope.cuda_init()

    return telescope

def test_ray_tracing_performance_large_mirror(benchmark, telescope_setup):
    telescope = telescope_setup
    source = iactsim.optics.sources.Source(telescope)
    source.positions.radial_uniformity = False
    source.positions.random = True
    source.positions.r_max = 11000
    source.set_target(distance=11000)

    def setup_round_on_axis():
        ps, vs, wls, ts = source.generate(1_000_000)
        cp.cuda.Device().synchronize()
        return ((telescope, ps, vs, wls, ts), {})

    def run_trace(telescope, ps, vs, wls, ts):
        telescope.trace_photons(ps, vs, wls, ts)
        cp.cuda.Device().synchronize()

    benchmark.pedantic(run_trace, setup=setup_round_on_axis, rounds=100, iterations=1, warmup_rounds=5)

def test_ray_tracing_performance_large_mirror_off_axis(benchmark, telescope_setup):
    telescope = telescope_setup
    source = iactsim.optics.sources.Source(telescope)
    source.positions.radial_uniformity = False
    source.positions.random = True
    source.positions.r_max = 11000
    source.directions.altitude += 1
    source.set_target(distance=11000)
    def setup_round_off_axis():
        ps, vs, wls, ts = source.generate(1_000_000)
        cp.cuda.Device().synchronize()
        return ((telescope, ps, vs, wls, ts), {})

    def run_trace(telescope, ps, vs, wls, ts):
        telescope.trace_photons(ps, vs, wls, ts)
        cp.cuda.Device().synchronize()

    benchmark.pedantic(run_trace, setup=setup_round_off_axis, rounds=100, iterations=1, warmup_rounds=5)


# Copyright (C) 2024- Davide Mollica <davide.mollica@inaf.it>
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of iactsim.
#
# iactsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# iactsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iactsim.  If not, see <https://www.gnu.org/licenses/>.

import pytest

import iactsim
import numpy as np
import cupy as cp

def dummy_pulse(extent, resolution, peak_delay, sigma):
    t = np.arange(0, extent, resolution)
    z = (t - peak_delay) / sigma
    log_norm = -0.5 * (z**2 + np.log(2*np.pi)) - np.log(sigma)
    y = np.where(log_norm<10, t**2*np.exp(log_norm), 0)

    wave = iactsim.electronics.Waveform(t, y, 'AC')
    wave.normalize()
    return wave
    
class DummySiPMCamera(iactsim.electronics.CherenkovSiPMCamera):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.images = []

    def restart(self):
        super().restart()
        self.images = []
    
    def sampling_action(self):
        image = cp.max(self.get_channel_signals(1, reshape=True), axis=1)
        self.images.append(image)
    
    def trigger_action(self):
        pass

@pytest.fixture
def cam_fast_dc_signals():
    waveforms = (
        dummy_pulse(50, 0.1, 0, 5),
        dummy_pulse(50, 0.1, 3, 8),
    )
    cam = DummySiPMCamera(
        n_pixels = 64*25,
        waveforms = waveforms,
        trigger_channels=[0],
        sampling_channels=[1],
        channels_time_resolution=[1,1]
    )
    return cam

@pytest.fixture
def cam_fast_dc_signal_4channels():
    wave = dummy_pulse(50, 0.1, 0, 5)

    waveforms = (
        wave,
        wave,
        wave,
        wave,
    )

    cam = DummySiPMCamera(
        n_pixels = 64*4,
        waveforms = waveforms,
        trigger_channels=[0,1],
        sampling_channels=[2,3],
        channels_time_resolution=[1,1,1,1]
    )
    cam.sampling_window_extent = [1024, 1024]
    cam.sampling_delay = [0,0]
    return cam

@pytest.mark.parametrize("xt", [0.0, 0.2])
@pytest.mark.parametrize("ap", [0.0, 0.2])
@pytest.mark.parametrize("n_ucells", [0, 1024])
def test_background_signal_channel_seed(cam_fast_dc_signal_4channels, xt, ap, n_ucells):
    """Test if all channels see the same background.
    I.e., the same background must be generated at each compute_signals() call.
    """
    cam = cam_fast_dc_signal_4channels

    cam.source = None
    cam.background_rate = cp.full((cam.n_pixels,), 0.5, dtype=cp.float32)
    cam.cross_talk = cp.full((cam.n_pixels,), xt, dtype=cp.float32)
    cam.afterpulse = cp.full((cam.n_pixels,), ap, dtype=cp.float32)
    cam.afterpulse_inv_tau = cp.full((cam.n_pixels,), 1./250., dtype=cp.float32)
    cam.inv_tau_recovery = cp.full((cam.n_pixels,), 1./200., dtype=cp.float32)

    cam.simulate_ucells = n_ucells > 0
    cam.n_ucells = cp.full((cam.n_pixels,), n_ucells, dtype=cp.int32)

    # Fixed windows
    cam.fixed_time_windows = True
    cam.time_windows = [
        cp.arange(0,2048,1, dtype=cp.float32),
        cp.arange(0,2048,1, dtype=cp.float32),
        cp.arange(0,2048,1, dtype=cp.float32),
        cp.arange(0,2048,1, dtype=cp.float32)
    ]
    cam.compute_signals([0])
    cam.compute_signals([1])
    cam.compute_signals([2,3])

    signals = {i: cam.get_channel_signals(i, reshape=True) for i in range(cam.n_channels)}

    for i in range(cam.n_channels-1):
        for j in range(cam.n_pixels):
            cp.testing.assert_allclose(signals[i][j], signals[i+1][j], atol=1e-6)

    # Compare pixels of each channel, they must be different
    channel = 0
    rng_weights = cp.random.randn(signals[channel].shape[1])
    fingerprints = cp.dot(signals[channel], rng_weights) 
    fingerprints.sort(axis=0)
    diffs = fingerprints[1:] - fingerprints[:-1]
    min_diff = float(cp.min(cp.abs(diffs)))
    assert min_diff > 1e-6

@pytest.mark.parametrize("ap", [0.0, 0.2])
@pytest.mark.parametrize("xt", [0.0, 0.2])
@pytest.mark.parametrize("n_ucells", [0, 1024])
def test_source_signal_channel_seed(cam_fast_dc_signal_4channels, xt, ap, n_ucells):
    """Test if all channels see the same background.
    I.e., the same background must be generated at each compute_signals() call.
    """
    cam = cam_fast_dc_signal_4channels

    # No background, xt and ap
    cam.background_rate = cp.full((cam.n_pixels,), 0.0, dtype=cp.float32)
    cam.cross_talk = cp.full((cam.n_pixels,), xt, dtype=cp.float32)
    cam.afterpulse = cp.full((cam.n_pixels,), ap, dtype=cp.float32)
    cam.afterpulse_inv_tau = cp.full((cam.n_pixels,), 1./250., dtype=cp.float32)
    cam.inv_tau_recovery = cp.full((cam.n_pixels,), 1./200., dtype=cp.float32)
    cam.sigma_ucells = cp.full((cam.n_pixels,), 0.1, dtype=cp.float32) # Big since when xt_and_ap is 0 this is the only source of randomness

    cam.n_ucells = cp.full((cam.n_pixels,), n_ucells, dtype=cp.int32)
    cam.simulate_ucells = n_ucells > 0

    # Same discharge times for all pixels
    t0s = cp.tile(cp.random.uniform(100.,200.,100), cam.n_pixels)
    map_ = cp.asarray([100*i for i in range(0,cam.n_pixels+1)])
    ucell_ids = cp.tile(cp.random.uniform(0,n_ucells,100), cam.n_pixels).astype(cp.int32)
   
    cam.source = (t0s, map_, ucell_ids) if cam.simulate_ucells else (t0s, map_)

    cam.base_seed = 0

    # Fixed windows
    cam.fixed_time_windows = True
    cam.time_windows = [
        cp.arange(0,2048,1, dtype=cp.float32),
        cp.arange(0,2048,1, dtype=cp.float32),
        cp.arange(0,2048,1, dtype=cp.float32),
        cp.arange(0,2048,1, dtype=cp.float32)
    ]
    cam.compute_signals([0])
    cam.compute_signals([1])
    cam.compute_signals([2,3])

    signals = {i: cam.get_channel_signals(i, reshape=True) for i in range(cam.n_channels)}

    # Compare channels -> they should be the same
    for i in range(cam.n_channels-1):
        for j in range(cam.n_channels-1):
            if i == j: continue
            cp.testing.assert_allclose(signals[i], signals[j], atol=1e-10)

    # Compare pixels of each channel, they must be different
    channel = 0
    rng_weights = cp.random.randn(signals[channel].shape[1])
    fingerprints = cp.dot(signals[channel], rng_weights) 
    fingerprints.sort(axis=0)
    diffs = fingerprints[1:] - fingerprints[:-1]
    min_diff = float(cp.min(cp.abs(diffs)))
    assert min_diff > 1e-6

@pytest.mark.parametrize("n_ucells", [0, 10000])
def test_background_signal_stats(cam_fast_dc_signals, n_ucells):
    """Test if the singal mean and the signal variance follow the Campbell's theorem. Cross-talk and asfterpulse are not included.
    Assuming that the waveforms are computed correctly, this test fails if the inter-pixel randomness is broken.
    """
    cam = cam_fast_dc_signals
    cam.timer.active = False
    cam.fixed_time_windows = True
    cam.time_windows = [
        cp.asarray([0], dtype=cp.float32),
        cp.arange(0,2048,1, dtype=cp.float32)
    ]
    cam.increment_seed = True
    cam.source = None
    cam.enable_camera_trigger = False
    cam.trigger_time = 0.
    cam.base_seed = 0

    # Simulate ucell with a very fast recovery time
    cam.inv_tau_recovery = cp.full((cam.n_pixels,), 1./10., dtype=cp.float32)
    cam.n_ucells = cp.full((cam.n_pixels,), n_ucells, dtype=cp.int32)
    cam.simulate_ucells = n_ucells > 0

    for background in [0.06, 0.1, 0.15]:
        cam.restart()
        cam.background_rate = cp.full((cam.n_pixels,), background, dtype=cp.float32)
        var = []
        mean = []
        for _ in range(100):
            cam.simulate_response()
            var.append(cp.var(cam.get_channel_signals(1)).get())
            mean.append(cp.mean(cam.get_channel_signals(1)).get())
        var = np.mean(var)
        mean = np.mean(mean)
        cp.testing.assert_allclose(cam.waveforms[1].get_squared_integral()*background, var, rtol=1e-3)
        cp.testing.assert_allclose(cam.waveforms[1].get_integral()*background, mean, rtol=1e-3)

def test_cross_talk_effect_on_background_signal(cam_fast_dc_signals):
    """Test if the singal mean and the signal variance follow the generalized Campbell's theorem.
    Cross-talk is included. Afterpuls is not included.
    Assuming that the waveforms are computed correctly, this test fails if the inter-pixel randomness is broken.
    """
    cam = cam_fast_dc_signals
    cam.base_seed = 0
    cam.timer.active = False
    cam.fixed_time_windows = True
    cam.time_windows = [
        cp.asarray([0], dtype=cp.float32),
        cp.arange(0,2048,1, dtype=cp.float32)
    ]
    cam.increment_seed = True
    cam.source = None
    cam.enable_camera_trigger = False
    cam.trigger_time = 0.

    background = 0.05
    cam.background_rate = cp.full((cam.n_pixels,), background, dtype=cp.float32)

    xts = np.asarray([0.0, 0.1, 0.2, 0.3])
    expected_var = cam.waveforms[1].get_squared_integral()*background/(1-xts)**3
    expected_mean = cam.waveforms[1].get_integral()*background/(1-xts)
    for i in range(xts.shape[0]):
        cam.restart()
        cam.cross_talk = cp.full((cam.n_pixels,), xts[i], dtype=cp.float32)
        var = []
        mean = []
        for _ in range(100):
            cam.simulate_response()
            var.append(cp.var(cam.get_channel_signals(1)).get())
            mean.append(cp.mean(cam.get_channel_signals(1)).get())
        var = np.mean(var)
        mean = np.mean(mean)
        cp.testing.assert_allclose(expected_mean[i], mean, rtol=1e-3)
        cp.testing.assert_allclose(expected_var[i], var, rtol=1e-3)


# Copyright (C) 2024- Davide Mollica <davide.mollica@inaf.it>
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of iactsim.
#
# iactsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# iactsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iactsim.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np

def pointing_dir(altitude: float, azimuth: float) -> np.ndarray:
    """Telescope pointing direction in the corsika-like local reference frame.

    *   North -> (0, 0) : x-axis
    *   East -> (0, 90)
    *   South -> (0, 180)
    *   West -> (0, 270) : y-axis
    *   Zenith -> (90, any) : z-axis

    Parameters
    ----------
    altitude : float
        Altitude angle in degrees.
    azimuth : float
        Azimuth angle in degrees. Measured from North, increasing towards East (clockwise).

    Returns
    -------
    direction : ndarray, shape (3,)
        Pointing unit vector in the local reference frame (North-East-Up).
    """
    azimuth_rad = np.deg2rad(azimuth)
    altitude_rad = np.deg2rad(altitude)
    return np.array([
        np.cos(altitude_rad) * np.cos(azimuth_rad),
        -np.cos(altitude_rad) * np.sin(azimuth_rad),
        np.sin(altitude_rad),
    ])

def local_to_pointing_rotation(altitude: float, azimuth: float) -> np.ndarray:
    """Compute the rotation matrix to transform from the local reference frame (North-West-Up) to the "pointing reference frame".

    This reference frame is defined such that when the telescope is
    pointing at the horizon towards the North (i.e. alt: 0, az: 0), the axes are aligned as follows:

    *   z-axis: aligned with the telescope optical axis, pointing North.
    *   x-axis: perpendicular to the optical axis, pointing downward.
    *   y-axis: perpendicular to the optical axis, pointing west.

    Parameters
    ----------
    altitude : float
        Altitude angle of the telescope pointing, in degrees.
    azimuth : float
        Azimuth angle of the telescope pointing, in degrees. Measured from North, increasing towards East.

    Returns
    -------
    R : ndarray, shape (3, 3)
        Rotation matrix. When a vector in the local reference frame is multiplied by this matrix
        it is transformed into the telescope reference frame.

    Notes
    -----
    This is a simple rotation of the CORSIKA reference frame that move the up-vector (z) into the telescope direction.

    """
    theta = np.radians(90-altitude)
    phi = -np.radians(azimuth)

    ct = np.cos(theta)
    st = np.sin(theta)
    cp = np.cos(phi)
    sp = np.sin(phi)

    # pointing north (0,0)
    #  z axis is pointing north (local x axis)
    #  x axis is pointing down (opposite to local z axis)
    #  y axis is pointing west (local y axis)
    R = np.array([
        [ct*cp, ct*sp, -st],
        [  -sp,    cp,   0],
        [st*cp, st*sp,  ct]
    ])
    return R

def local_to_telescope_rotation(altitude: float, azimuth: float, custom_rotation=None) -> np.ndarray:
    """Compute the rotation matrix to transform from the local reference frame (North-West-Up) to the "telescope reference frame".
    By deault the telescope refrence frame is the pointing reference frame (see :py:func:`local_to_pointing_rotation`).

    Parameters
    ----------
    altitude : float
        Altitude angle of the telescope pointing, in degrees.
    azimuth : float
        Azimuth angle of the telescope pointing, in degrees. Measured from North, increasing towards East.

    Returns
    -------
    R : ndarray, shape (3, 3)
        Rotation matrix. When a vector in the local reference frame is multiplied by this matrix
        it is transformed into the telescope reference frame.

    Examples
    --------
    To rotate the telescope frame around the optical axis (Z-axis), while keeping Z pointing in the same direction:

        .. code-block:: python

            import numpy as np

            # Rotation around Z-axis
            rotation = np.array([
                [ 0, 1, 0], # new X is old Y
                [ 1, 0, 0], # new Y is old X
                [ 0, 0, 1] 
            ])

            # Get the telescope transformation matrix
            R_tel = local_to_telescope_rotation(
                altitude=70, 
                azimuth=180, 
                custom_rotation=rotation
            )
    
    Now when the telescope is pointing North:

    *   z-axis: aligned with the telescope optical axis, pointing North;
    *   y-axis: perpendicular to the optical axis, pointing downward;
    *   x-axis: perpendicular to the optical axis, pointing west.

    """
    # Rotation into the local reference frame
    R = local_to_pointing_rotation(altitude, azimuth)
    
    if custom_rotation is None:
        custom_rotation = np.eye(3)

    return custom_rotation @ R

def local_to_photon_rotation(u: np.ndarray) -> np.ndarray:
    """Compute the rotation matrix to transform vector `u` to the local vertical Z-axis (0,0,1).
    
    This uses the Rodrigues rotation formula.

    Parameters
    ----------
    u : ndarray, shape (3,)
        The source unit vector (e.g., photon direction).
    
    Returns
    -------
    R : ndarray, shape (3, 3)
        Rotation matrix such that R @ u = [0, 0, 1].
    """
    # Normalize input
    u = u / np.linalg.norm(u)
    
    # Define target (transform in a reference system where u = v)
    v = np.array([0.0, 0.0, 1.0])

    # Cosine of the angle between u and v
    c = np.dot(u, v)
    
    # Vectors are already parallel
    if c > 0.999999:
        return np.eye(3)
        
    # Vectors are anti-parallel
    # Rotate 180 degrees around any axis perpendicular to u.
    if c < -0.999999:
        # Pick a perpendicular axis
        # If u is not parallel to X, use cross(u, X). Else use cross(u, Y).
        if np.abs(u[0]) < 0.9:
            axis = np.cross(u, np.array([1, 0, 0]))
        else:
            axis = np.cross(u, np.array([0, 1, 0]))
        axis = axis / np.linalg.norm(axis)
        
        # Rodrigues formula for 180 degrees
        #   R = 2(axis * axis.T) - I
        return 2 * np.outer(axis, axis) - np.eye(3)

    ### Rodrigues rotation formula
    #   R = I + K + K^2 * (1 / (1 + c))
    # See for example https://math.stackexchange.com/a/476311
    
    # Rotation axis k = u x v
    k = np.cross(u, v)
    
    # Skew-symmetric cross-product matrix of k
    # [ 0  -k.z  k.y]
    # [ k.z  0  -k.x]
    # [-k.y  k.x  0 ]
    K = np.array([
        [   0., -k[2],  k[1]],
        [ k[2],    0., -k[0]],
        [-k[1],  k[0],    0.]
    ])
    
    R = np.eye(3) + K + (K @ K) * (1. / (1. + c))
    
    return R

def create_surface_rotation(normal: np.ndarray, up: np.ndarray = np.array([1, 0, 0])) -> np.ndarray:
    """
    Computes a rotation matrix to transform from the telescope reference frame to the surface reference frame.

    In the surface reference system:

    * the Z-axis (0, 0, 1) is aligned with the input ``normal``;
    * the Y-axis (0, 1, 0) is aligned as close as possible to the input ``up`` vector;
    * the X-axis (1, 0, 0) is mutually orthogonal.

    Parameters
    ----------
    normal : np.ndarray
        The vector in the telescope reference system that defines where the surface is facing. This will become the local Z-axis.
    up : np.ndarray, optional
        The vector in the telescope reference system that defines the "top" of the surface. 
        The local Y-axis will be constructed by projecting this vector onto the plane 
        perpendicular to the normal. Defaults to the X axis of telescope reference frame ([1, 0, 0]).

    Returns
    -------
    R : ndarray, shape (3, 3)
        The 3x3 rotation matrix. 
        
        * Usage: 
          
           - ``v_surf = R @ v_telescope`` if ``v_telescope`` has shape (3,)
           - ``v_surf = v_telescope @ R.T`` if ``v_telescope`` has shape (N,3)

    Notes
    -----

    The ID of each pixel of a SiPM tile is defined in the tile reference system as follows::

        ^ y
        |     _____   _____   _____
        |    |     | |     | |     |
        |    |  6  | |  7  | |  8  |
        |    |_____| |_____| |_____|
        |     _____   _____   _____
        |    |     | |     | |     |
        |    |  3  | |  4  | |  5  |
        |    |_____| |_____| |_____|
        |     _____   _____   _____
        |    |     | |     | |     |
        |    |  0  | |  1  | |  2  |
        |    |_____| |_____| |_____|
        |
        ------------------------------> x

    So, the SiPM tile rotation matrix must be carefully set to match the desired ID convention.
    
    Examples
    --------

        .. code-block:: python

            # Opposite to the pointing direction
            sensor_normal = np.array([0., 0., -1.])

            # Align the surface y axis with the telescope y axis
            R = create_surface_rotation(normal=sensor_normal, up=np.array([0., 1., 0.]))
        
    """
    # Surface Z axis is the normal
    z_axis = normal / np.linalg.norm(normal)
    
    # New X axis is perpendicular to both up and normal vectors
    x_axis = np.cross(up, z_axis)
    
    # Handle collinearity (if normal is parallel to up)
    if np.linalg.norm(x_axis) < 1e-6:
        raise(ValueError("Up vector and normal vector are parallel."))
    
    x_axis = x_axis / np.linalg.norm(x_axis)
    
    # New Y axis is perpendicular to Z and X
    y_axis = np.cross(z_axis, x_axis)
    
    # Construct rotation matrix
    R = np.array([
        x_axis,
        y_axis,
        z_axis
    ])
    
    return R


# Copyright (C) 2024- Davide Mollica <davide.mollica@inaf.it>
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of iactsim.
#
# iactsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# iactsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iactsim.  If not, see <https://www.gnu.org/licenses/>.

from typing import Optional, TYPE_CHECKING
if TYPE_CHECKING:
    from ..._iact import IACT

import numpy as np
import cupy as cp
import matplotlib.pyplot as plt

from .directions import (
    AbstractPhotonDirections,
    PointLike as PointLikeDir,
    UniformBeam,
    GaussianBeam
)

from .positions import (
    AbstractPhotonPositions,
    PointLike as PointLikePos,
    UniformDisk,
    UniformSphericalCap
)

from .._cpu_transforms import pointing_dir, local_to_pointing_rotation, local_to_telescope_rotation
from .._surface import SurfaceType

class Spectrum():
    """Generate a photon spectral distribution.

    Parameters
    ----------
    type : str
        Distribution type. Supported values are: "uniform", "constant", and "cherenkov".
    
    """
    
    __spectra = ['uniform', 'constant', 'cherenkov']

    def __init__(self, type: str):
        self.type: str = type  #: Distribution type, supported values are: "uniform", "constant" and "cherenkov".
        self.lmin: float = 200.  #: Minimum wavelength (in nm).
        self.lmax: float = 900.  #: Maximum wavelength (in nm).
        self.l0: float = 450.  #: Central wavelength (in nm, used only if :attr:`type` is "constant").
        self.dtype: type = cp.float32  #: Data type for generated wavelengths.
        self.seed = None

    @property
    def seed(self):
        return self._seed
    
    @seed.setter
    def seed(self, a_seed):
        self.rng = cp.random.default_rng(cp.random.Philox4x3210(a_seed))
        self._seed = a_seed
    
    @property
    def type(self) -> str:
        """
        str
            The distribution type ("uniform", "constant", or "cherenkov")
        """
        return self.__type

    @type.setter
    def type(self, dist_type: str):
        if dist_type not in self.__spectra:
            raise ValueError(f"Spectrum '{dist_type}' not supported. Supported choices are {self.__spectra}")
        self.__type = dist_type

    def generate(self, n: int) -> cp.ndarray:
        """
        Generates wavelength samples based on the specified distribution.

        Parameters
        ----------
        n : int
            The number of samples to generate.

        Returns
        -------
        sample : ndarray
            An array of generated wavelength samples.

        Raises
        ------
        NotImplementedError
            If the distribution type is not recognized.
        """
        if self.__type == 'cherenkov':
            emin = 1. / self.lmax
            emax = 1. / self.lmin
            sample = 1. / self.rng.uniform(emin, emax, size=n, dtype=self.dtype)

        elif self.__type == 'uniform':
            sample = self.rng.uniform(self.lmin, self.lmax, size=n, dtype=self.dtype)
        elif self.__type == 'constant':
            sample = cp.full((n,), self.l0, dtype=self.dtype)
        else:
            raise NotImplementedError(f"Distribution {self.__type} is not yet implemented.")
        return sample

class TimeDistribution():
    """Generate photon arrival times.
        
        Parameters
        ----------
        type : str
            Distribution type. Supported values are: "uniform" (i.e. Poissonian) and "constant".

    """
    __dists = ['uniform', 'constant']

    def __init__(self, type: str):
        self.type: str = type  #: Distribution type, supported values are: "uniform" and "constant".
        self.tmin: float = 0.  #: Start time (in ns).
        self.tmax: float = 10.  #: End time (in ns).
        self.t0: float = 0.  #: Central time (in ns, used only if :attr:`type` is "constant").
        self.dtype: type = cp.float32  #: Data type for generated times.
        self.seed = None

    @property
    def seed(self):
        return self._seed
    
    @seed.setter
    def seed(self, a_seed):
        self.rng = cp.random.default_rng(cp.random.Philox4x3210(a_seed))
        self._seed = a_seed
    
    @property
    def type(self) -> str:
        """
        str
            The distribution type ("uniform" or "constant").
        """
        return self.__type

    @type.setter
    def type(self, dist_type: str):
        if dist_type not in self.__dists:
            raise ValueError(f"Time distribution '{dist_type}' not supported. Supported choices are {self.__dists}")
        self.__type = dist_type

    def generate(self, n: int) -> cp.ndarray:
        """
        Generates time samples based on the specified distribution.

        Parameters
        ----------
        n : int
            The number of samples to generate.

        Returns
        -------
        sample : ndarray
            An array of generated time samples.

        Raises
        ------
        NotImplementedError
            If the distribution type is not recognized.
        """
        if self.__type == 'uniform':
            sample = self.rng.uniform(self.tmin, self.tmax, size=n, dtype=self.dtype)
        elif self.__type == 'constant':
            sample = cp.full((n,), self.t0, dtype=self.dtype)
        else:
            raise NotImplementedError(f"Distribution {self.__type} is not yet implemented.")
        return sample

class Source():
    """Represents a photon source.
    By default a point-like source at zenith is generated. Photons are generated at a point (0, 0, 100) m in the local reference frame. 

    Parameters
    ----------
    telescope : IACT | None, optional
        The IACT object associated with this source. If None, the source is assumed to be at a fixed position defined by `self.positions.position`.
        Otherwise, a point-like source is generated as a uniform disk, at a coordinates (0, 0, 100) m in the local reference frame, with a radius big enough to illimunate the aperture of the telescope.
        
    dtype : type, optional
        The desired data type for generated quantities, by default np.float32.
    """
    def __init__(self, telescope: Optional['IACT'] = None, dtype: type = np.float32):
        self._tel: 'IACT' | None = telescope

        self.spectrum: Spectrum = Spectrum('uniform')  #: Photon wavelengths generator.
        self.spectrum.dtype = dtype

        self.arrival_times: TimeDistribution = TimeDistribution('constant')  #: Photon arrival times generator.
        self.arrival_times.dtype = dtype
        
        self.positions: AbstractPhotonPositions = PointLikePos([0,0,1e5]) #: Photon positions generator.

        self.directions: AbstractPhotonDirections = PointLikeDir(90,0) #: Photon directions generator.

        if telescope is not None:
            # On axis source by default
            self.directions.altitude = telescope.pointing[0]
            self.directions.azimuth = telescope.pointing[1]
            # Uniform disk by default
            radii = [s.half_aperture for s in self._tel.optical_system if hasattr(s, 'half_aperture')]
            # Relative distance 
            radial_position = [np.sqrt(s.position[0]**2+s.position[1]**2) for s in self._tel.optical_system if hasattr(s, 'position')]
            rmax = 2.*max(radii) if len(radii)>0 else 4000.
            if len(radial_position)>0:
                rmax += 2.*max(radial_position) 
            self.positions = UniformDisk(0, rmax)
            self.set_target()
    
    
    def set_target(self, target_surface: str | int | None = None, distance: float = 1e5):
        """Adjust source position to match the center of a target surface.
        If the target surface is not specified it is assumed to be the telescope position.

        Parameters
        ----------
        target_surface : int or str, optional
            Name or index of the surface in the telescope optical system, by default the telescope position is assumed as the target.
        distance : float
            Source-target distance in mm.
        """
        if self._tel is None:
            raise(RuntimeError('Source must be linked to a telescope when initialized in order to set a target.'))
        
        if target_surface is None:
            target = self._tel
        else:
            target = self._tel.optical_system[target_surface]
        
        source_central_dir = self.directions.altitude, self.directions.azimuth

        # Set source origin position in order to point the target
        
        # Get the position of the target in the local reference frame
        tel_to_local_rotation = self._tel.local_to_telescope_rotation.T
        p = tel_to_local_rotation @ target.position + self._tel.position
        p = cp.asarray(p)

        # Get the direction of the source
        central_dir = pointing_dir(*source_central_dir)

        # Center the source on the target
        p0 = p + cp.asarray(distance*central_dir)
        self.positions.position = p0

        # Rotate position into the source direction
        # N.B. extracted position are rotated and then translated into p0
        self.positions.rotation = local_to_pointing_rotation(self.directions.altitude, self.directions.azimuth).T

    def generate(self, n: int) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Generate *n* photons.

        Parameters
        ----------
        n : int
            Number of photons to generate.

        Returns
        -------
        ps, vs, wls, ts: tuple of ndarrays
            Position, direction, wavelength and arrival time of photons.
        """
        
        ps = self.positions.generate(n)
        # Non-random source can generate a lower number of photons
        if ps.shape[0] != n:
            n = ps.shape[0]
        vs = self.directions.generate(n)
        wls = self.spectrum.generate(n)
        ts = self.arrival_times.generate(n)
        return ps, vs, wls, ts
    
    def visualize(self, n_photons: int = 1000, opacity=None, length=None, show_hits=True, show_rays=True, resolution=None):
        """VTK visualization of the source.

        Parameters
        ----------
        n_photons : int, optional
            Number of photons to generate, by default 1000.
        opacity : float, optional
            Ray opacity, by default None (auto).
        length : float, optional
            If provided, rays that not intersects the telescope will be drawn with this length using the direction vector.
        show_hits : bool, optional
            Whether to show hits, by default True.
        show_rays : bool, optional
            Whether to show rays, by default True.
        resolution : float, optional
            Objects mesh resolution (in mm). By default None (depends on object extent).
        
        Raises
        ------
        RuntimeError
            If the source is not linked to a telescope.
        """
        # Import vtk here, so we are not force to add it to the package requirements
        from ...visualization._vtk_optical_system import VTKOpticalSystem

        if self._tel is None:
            raise RuntimeError('Source must be linked to a telescope when initialized in order to visualize it.')
        
        # Transform every surface into a sensitive one
        old_types = []
        old_props = []
        for s in self._tel.optical_system:
            old_types.append(s.type)
            old_props.append(s.properties)
            s.type = SurfaceType.SENSITIVE
            s.properties = None
        self._tel.cuda_init()

        # Generate photons and trace them into the telescope
        ps, vs, wls, ts = self.generate(n_photons)
        ps_start = ps.get()
        vs_start = vs.get()
        self._tel.trace_photons(
            ps, vs, wls, ts,
            save_last_bounce=True
        )
        ps_stop = ps.get()

        # Get the surface ID
        surface_hits = (self._tel._d_weights >> 16) & 0xFFFF
        surface_hits = surface_hits.astype(cp.int32).get() - 1

        # VTK renderer
        renderer = VTKOpticalSystem(self._tel, resolution=resolution)
        renderer.add_rays(
            ps_start, ps_stop,
            surface_id=surface_hits,
            length=length,
            directions=vs_start,
            show_hits=show_hits,
            show_rays=show_rays,
        )
        renderer.set_ray_opacity(opacity)
        renderer.start_render()

        # Reset the surface types
        for s,s_type,s_prop in zip(self._tel.optical_system, old_types, old_props):
            s.type = s_type
            s.properties = s_prop
        self._tel.cuda_init()


# Copyright (C) 2024- Davide Mollica <davide.mollica@inaf.it>
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of iactsim.
#
# iactsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# iactsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iactsim.  If not, see <https://www.gnu.org/licenses/>.

import cupy as cp
import numpy as np
from numpy.typing import NDArray
from dataclasses import dataclass, field
from tabulate import tabulate
from typing import Optional, Iterator, Dict, ClassVar, Union
import keyword

from ._air_refractive_index import calculate_ciddor_rindex

def calculate_ciddor_1996_std_air(wavelength_nm: NDArray[np.float64]) -> NDArray[np.float64]:
    """Calculates the refractive index of standard air using the Ciddor 1996 equation.
    Standard conditions: 15 °C, 101325 Pa, 0% RH, 400 ppm CO2.
    Valid for vacuum wavelength from 230 nm to 1690 nm.

    Parameters
    ----------
    wavelength_nm : NDArray[np.float64]
        Wavelength (in nm).

    Returns
    -------
    NDArray[np.float64]
        Refractive index.
    """

    # um^-2 constants
    k_0 = 238.0185
    k_1 = 5792105
    k_2 = 57.362
    k_3 = 167917
    wavelength_um = wavelength_nm / 1000.0
    inv_wavelength_um_sq = (1.0 / wavelength_um)**2

    refractivity = (k_1 / (k_0 - inv_wavelength_um_sq)) + (k_3 / (k_2 - inv_wavelength_um_sq))
    n = refractivity/1e8 + 1.
    return n

def calculate_fused_silica_sellmeier(wavelength_nm: NDArray[np.float64]) -> NDArray[np.float64]:
    """
    Calculates the refractive index of Fused Silica using the Sellmeier
    equation with coefficients from Malitson (1965).
    Valid for vacuum wavelength from 0.21 um to 3.71 um.

    Parameters
    ----------
    wavelength_nm : NDArray[np.float64]
        Wavelength (in nm).

    Returns
    -------
    NDArray[np.float64]
        Refractive index.
    """
    # Sellmeier coefficients from Malitson (1965)
    B1 = 0.6961663
    C1 = 0.0684043**2 # um^2
    B2 = 0.4079426
    C2 = 0.1162414**2 # um^2
    B3 = 0.8974794
    C3 = 9.896161**2  # um^2

    # Convert wavelength from nm to um
    wavelength_um = wavelength_nm / 1000.0
    wavelength_um_sq = wavelength_um**2

    # Calculate n^2 - 1
    term1 = B1 * wavelength_um_sq / (wavelength_um_sq - C1)
    term2 = B2 * wavelength_um_sq / (wavelength_um_sq - C2)
    term3 = B3 * wavelength_um_sq / (wavelength_um_sq - C3)
    n_sq_minus_1 = term1 + term2 + term3

    n_sq = n_sq_minus_1 + 1.0
    n = np.sqrt(np.maximum(0, n_sq))
    return n

@dataclass
class RefractiveIndexTexture:
    """
    Holds the texture object and metadata for refractive index lookups.
    It retains a reference to the underlying CUDA array to prevent
    premature garbage collection.
    """
    texture: cp.cuda.texture.TextureObject
    wavelength_start: float
    wavelength_inv_step: float
    
    _cuda_array: cp.cuda.texture.CUDAarray = field(default=None, repr=False)

@dataclass(frozen=True)
class Material:
    name: str
    value: int
    wavelength: Optional[NDArray[np.float64]] = field(default=None, repr=False) # nm
    n: Optional[NDArray[np.float64]] = field(default=None, repr=False)
    
    def __repr__(self):
        return f"<Material: {self.name} (value={self.value})>"
        
    def get_refractive_index(self, wavelength_nm: Union[float, np.ndarray]) -> Union[float, np.ndarray, None]:
        if self.wavelength is None or self.n is None:
            return None

        result = np.interp(wavelength_nm, self.wavelength, self.n, left=np.nan, right=np.nan)
        
        # If input is a scalar and the result is NaN (out of bounds), return None.
        if np.isscalar(wavelength_nm) and np.isnan(result):
            return None

        return result
    
    def get_refractive_index_texture(self):
        # Texture Setup
        width = len(self.wavelength)
        texture_data = self.n.astype(np.float32)
        
        # Create a CUDAarray for the 1D texture
        desc = cp.cuda.texture.ChannelFormatDescriptor(32, 0, 0, 0, cp.cuda.runtime.cudaChannelFormatKindFloat)
        cu_array = cp.cuda.texture.CUDAarray(desc, width)
        cu_array.copy_from(texture_data)
        
        # Create another texture object (Linear Interpolation)
        res_desc_linear = cp.cuda.texture.ResourceDescriptor(
        cp.cuda.runtime.cudaResourceTypeArray,
        cuArr=cu_array
        )
        tex_desc_linear = cp.cuda.texture.TextureDescriptor(
            addressModes = (cp.cuda.runtime.cudaAddressModeClamp,),
            filterMode = cp.cuda.runtime.cudaFilterModeLinear,
            readMode = cp.cuda.runtime.cudaReadModeElementType,
            normalizedCoords = False
        )

        wl_start = self.wavelength[0]
        wl_inv_step = 1./(self.wavelength[1]-self.wavelength[0])
        tex_obj = cp.cuda.texture.TextureObject(res_desc_linear, tex_desc_linear)

        return RefractiveIndexTexture(
            texture=tex_obj,
            wavelength_start=wl_start,
            wavelength_inv_step=wl_inv_step,
            _cuda_array=cu_array
        )

class MaterialsMeta(type):
    """
    Metaclass for the Materials class.

    Manages the registration of materials and dynamically adds registered materials as class attributes.
    Material uniqueness and overwriting are based on the standardized attribute name (e.g., 'AIR').
    """
    _registry: ClassVar[Dict[int, Material]] = {} # Maps integer value -> Material
    _name_registry: ClassVar[Dict[str, Material]] = {} # Maps standardized name string -> Material
    _next_value: ClassVar[int] = 0 # Counter for assigning unique values

    def __str__(cls):
        if not cls._registry:
            return "<Materials: (No materials registered)>"
        return ", ".join(str(material) for material in cls._members())

    def __repr__(cls):
        if not cls._registry:
            return "<Materials: (No materials registered)>"
        return f"<Materials: {', '.join(repr(material) for material in cls._members())}>"

    def _repr_html_(cls):
        if not cls._registry:
            return "<p><b>Materials:</b> (No materials registered)</p>"
        return cls._to_table(tablefmt="html")

    def _repr_markdown_(cls):
        if not cls._registry:
            return "**Materials:** (No materials registered)"
        return cls._to_table(tablefmt="pipe")

    def _to_table(cls, tablefmt="html") -> str:
        table_data = [[material.name, material.value] for material in cls._members()]
        return tabulate(table_data, headers=["Material Name (Standardized)", "Key"], tablefmt=tablefmt)
    
    def _members(cls) -> Iterator[Material]:
        """Returns an iterator over registered Material instances, sorted by value."""
        return iter(sorted(cls._registry.values(), key=lambda m: m.value))

    def register_material(cls,
                          input_name: str,
                          wavelengths_nm: NDArray[np.float64],
                          n: NDArray[np.float64],
                          overwrite: bool = False) -> Material:
        """
        Registers a new material or updates an existing one based on its
        standardized attribute name (e.g., 'AIR', 'FUSED_SILICA').

        The standardized name (uppercase, underscores) is used for uniqueness checks 
        and as the `name` attribute of the created Material object.
        If overwrite=True, it replaces the material with the same standardized name
        in the registry and updates the corresponding class attribute.

        Parameters
        ----------
        input_name : str
            The input name for the material (e.g., "Air", "Fused Silica").
            This will be standardized internally.
        wavelengths_nm : NDArray[np.float64]
            1D NumPy array of wavelengths in nanometers. Must be sorted ascending
            with uniform spacing.
        n : NDArray[np.float64]
            1D NumPy array of refractive indices corresponding to wavelengths_nm.
            Must have the same shape as wavelengths_nm.
        overwrite : bool, optional
            If True, allows overwriting an existing material with the same
            standardized name. Defaults to False.

        Returns
        -------
        Material
            The registered or updated Material instance.

        Raises
        ------
        TypeError
            If wavelengths_nm or n are not 1D NumPy arrays.
        ValueError
            If shapes mismatch, less than two data points are provided,
            wavelengths are not sorted ascending, wavelengths do not have
            uniform spacing or a valid attribute name cannot be generated.
        KeyError
            If a material with the same standardized name exists and overwrite is False.
        """
        standardized_name = cls._generate_attribute_name(input_name)

        if not isinstance(wavelengths_nm, np.ndarray) or not isinstance(n, np.ndarray):
            raise TypeError("Inputs 'wavelengths_nm' and 'n' must be NumPy arrays.")
        
        if wavelengths_nm.ndim != 1 or n.ndim != 1:
            raise ValueError("Inputs 'wavelengths_nm' and 'n' must be 1D.")
        
        if wavelengths_nm.shape != n.shape:
            raise ValueError(f"Shape mismatch: {wavelengths_nm.shape} vs {n.shape}.")
        
        if len(wavelengths_nm) < 2:
            raise ValueError("Need at least two data points for interpolation.")
        
        diffs = np.diff(wavelengths_nm)
        if not np.all(diffs > 0):
            raise ValueError("Wavelengths must be sorted in strictly ascending order.")
        
        if len(diffs) > 0 and not np.allclose(diffs, diffs[0], rtol=1e-4, atol=1e-3):
             raise ValueError(f"Wavelengths should be uniformly spaced.")

        existing_material = cls._name_registry.get(standardized_name)
        value = None

        if existing_material:
            if not overwrite:
                raise KeyError(f"Material with standardized name '{standardized_name}' (derived from '{input_name}') already exists (value={existing_material.value}). Use overwrite=True to replace.")
            else:
                value = existing_material.value
        else:
            value = cls._next_value
            cls._next_value += 1

        new_material = Material(
            name=standardized_name,
            value=value,
            wavelength=wavelengths_nm.copy(),
            n=n.copy()
        )

        cls._registry[value] = new_material
        cls._name_registry[standardized_name] = new_material

        setattr(cls, standardized_name, new_material)

        return new_material

    @staticmethod
    def _generate_attribute_name(name: str) -> str:
        """
        Generates a valid Python attribute name from a material name.
        Converts to uppercase, replaces spaces/hyphens with underscores,
        removes invalid characters, ensures it doesn't start with a digit,
        and avoids Python keywords.
        """
        if not isinstance(name, str) or not name:
             raise ValueError("Input name must be a non-empty string.")
        
        # Convert to uppercase and replace common separators
        attr_name = name.upper().replace(' ', '_').replace('-', '_')

        # Keep only alphanumeric characters and underscores
        attr_name = ''.join(c for c in attr_name if c.isalnum() or c == '_')

        # Remove leading/trailing underscores that might result from replacements
        attr_name = attr_name.strip('_')

        if not attr_name:
            raise ValueError(f"Cannot generate a valid attribute name from '{name}' after cleaning.")
        
        # Ensure it doesn't start with a digit
        if attr_name[0].isdigit():
            attr_name = '_' + attr_name
        
        # Check again just in case strip resulted in empty or invalid start
        if not (attr_name[0].isalpha() or attr_name[0] == '_'):
             raise ValueError(f"Cannot generate a valid attribute name from '{name}' (invalid start '{attr_name[0]}').")
        
        # Avoid collision with Python reserved keywords
        if keyword.iskeyword(attr_name):
            attr_name += '_'
        
        return attr_name
    
class Materials(metaclass=MaterialsMeta):
    """
    A collection of Material instances with refractive index data (wavelength in nm).
    Includes default materials: Air (Ciddor 1996) and Fused Silica (Malitson 1965).

    New materials can be added using the ``register_material()`` method:

    .. code-block::python

        Materials.register_material("MyAir", my_wavelength, my_n)
        print(Materials)
    
    You can also replace a material, for example the air refractive index:

    .. code-block::python

        Materials.register_material("Air", my_wavelength, my_n, overwrite=True)
        print(Materials)
    
    To register a refractive index for air, a ``set_atmospheric_conditions()`` method is provied.
    It uses the Ciddor equation following the NIST implementation:

    .. code-block::python

        Materials.set_atmospheric_condition(
            76.38e3, # pressure (Pa)
            20., # air temperatur (degrees Celsius)
            450., # CO2 concentration (ppm)
            10., # relative humidity (%)
            lambda_min=230., # lambda range (in nm)
            lambda_max=1700. #
        )
    
    """
    @classmethod
    def set_atmospheric_conditions(
        cls,
        pressure: float = 101325.,
        temperature: float = 20.,
        co2_concentration: float = 450.,
        relative_humidity: Optional[float] = 0.,
        lambda_min: float = 200.,
        lambda_max: float = 1700.,
    ) -> None:
        """Calculate the refractive index of moist air using the Ciddor equation and update the default air material.

        Follows the NIST implementation described by `Jack A. Stone and
        Jay H. Zimmerman <https://emtoolbox.nist.gov/Wavelength/Documentation.asp>`_.

        Parameters
        ----------
        pressure : float
            Air pressure in Pascals.
        temperature : float
            Air temperature in degrees Celsius.
        co2_concentration : float
            CO2 concentration in micromoles per mole (µmol/mol), equivalent to ppm.
        relative_humidity : Optional[float]
            Relative humidity in percent (%). If None, assumes dry air (xv = 0).
        lambda_min: float, Optional[float]
            Minimum photon wavelength (in nm).
        lambda_max: float, Optional[float]
            Maximum photon wavelength (in nm).

        Notes
        -----
            NIST implementation is valid in the range 300 nm - 1700 nm.

        """
        _wl_air_nm = np.arange(lambda_min, lambda_max+1., 1., dtype=np.float64)
        _n_air = calculate_ciddor_rindex(
            lambda_vac=_wl_air_nm,
            p=pressure,
            t=temperature,
            xCO2=co2_concentration,
            rh=relative_humidity
        )
        cls.register_material("Air", _wl_air_nm, _n_air, overwrite=True)

# Air (using Ciddor 1996 for standard air, 230 nm - 1690 nm)
_wl_air_nm = np.arange(200., 1690.+1., 1., dtype=np.float64)
_n_air = calculate_ciddor_1996_std_air(_wl_air_nm)

# Fused Silica (using Malitson 1965, 210 nm - 3710 nm)
_wl_silica_nm = np.arange(200., 3710.+1., 1., dtype=np.float64)
_n_silica = calculate_fused_silica_sellmeier(_wl_silica_nm)

Materials.register_material("Air", _wl_air_nm, _n_air)
Materials.register_material("Fused Silica", _wl_silica_nm, _n_silica)


# Copyright (C) 2024- Davide Mollica <davide.mollica@inaf.it>
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of iactsim.
#
# iactsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# iactsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iactsim.  If not, see <https://www.gnu.org/licenses/>.

import re
from functools import wraps
import textwrap

def wrap_cupy_rawkernel(raw_kernel, docstring):
    """
    Generates wrapper function for a CuPy RawKernel.
    The wrapper allows you to provide a custom docstring.

    Parameters
    ----------
        raw_kernel: cupy.RawKernel
            The cupy.RawKernel object to wrap.
        docstring: str
            A custom docstring for the wrapped kernel.

    Returns
    -------
        A Python function that wraps the CuPy RawKernel.
    """

    @wraps(raw_kernel)  # Preserve original kernel's metadata
    def wrapper(*args, **kwargs):
        raw_kernel(*args, **kwargs)

    wrapper.__doc__ = docstring

    return wrapper

def clean(string):
    string_clean = re.sub(r"^ \* ?", "", string, flags=re.MULTILINE).strip()
    if string_clean.startswith("*"):
        string_clean = string_clean[1:]
    return string_clean

def convert_doxygen_to_numpy(cuda_source, function_name):
    """
    Converts Doxygen-style documentation within the given source code to NumPy-style docstrings.

    Parameters
    ----------
        cuda_source: str
            The source code.
        function_name: str
            The name of the function.

    Returns
    -------
        docstring: str 
            The NumPy-style docstring or None if not found.
    """
    function_pattern = re.compile(
        r"""
        /\*\*\s*\n\s*\*\s*@brief\s*(?P<doc_brief>.*?)\n(?P<doc_details>.*?)\*/
        \s*
        (?:
            (?P<cuda_qualifier>__global__|__device__)\s+[^{;]*\b(?P<func_name>\w+)\s*\(
            |
            (?P<struct_type>struct|class)\s+(?P<struct_name>\w+)\s*\{
        )
        """,
        re.DOTALL | re.VERBOSE
    )

    for match in function_pattern.finditer(cuda_source):

        current_function_name = match.group("func_name") or match.group("struct_name")
        if current_function_name != function_name:
            continue
        
        is_global = match.group("cuda_qualifier") == '__global__'

        start = match.start()
        open_paren_index = match.end()
        brace_count = 1
        cur_pos = open_paren_index
        while brace_count > 0 and cur_pos < len(cuda_source):
            if cuda_source[cur_pos] in ['(', '{']:
                brace_count += 1
            elif cuda_source[cur_pos] in [')', '}']:
                brace_count -= 1
            cur_pos += 1
        end = cur_pos
        function_source = cuda_source[start:end]
        function_source = re.sub(r'//.*', '', function_source)
        
        param_string_start = [m.start() for m in re.finditer(r'\(',function_source)][-1]+1
        param_string_end = [m.start() for m in re.finditer(r'\)',function_source)][-1]
        param_types_str = function_source[param_string_start:param_string_end].strip()
        param_types_str = re.sub(r"[\n\t ]+", " ", param_types_str)
        
        param_types_dict = {}
        if param_types_str:
            params = [p.strip() for p in param_types_str.split(",") if p.strip()]
            for param in params:
                param = re.sub(r'//.*', '', param).strip()
                if not param: continue
                parts = param.rsplit(" ", 1)
                if len(parts) == 2:
                    param_type, param_name = parts[0].strip(), parts[1].strip()
                    param_type = param_type.replace(' *', '*').replace(' &', '&')
                    if param_name.startswith('*') or param_name.startswith('&'):
                        param_type += param_name[0]
                        param_name = param_name[1:]
                    param_types_dict[param_name] = param_type
                elif len(parts) == 1:
                    param_types_dict[parts[0]] = "Unknown"

        brief = match.group("doc_brief").strip()
        detailed = match.group("doc_details").strip()

        param_pattern = re.compile(r"@param\s+(\w+)\s+((?:(?!@param|@warning|@note|@return|@returns|@see|@code|@details).)+)", re.DOTALL)
        params = param_pattern.findall(detailed)
        warnings = re.findall(r'@warning\s+((?:(?!@param|@warning|@note|@return|@returns|@see|@code|@details).)+)', detailed, re.DOTALL)
        notes = re.findall(r'@note\s+((?:(?!@param|@warning|@note|@return|@returns|@see|@code|@details).)+)', detailed, re.DOTALL)
        returns = re.findall(r'@return(?:s)?\s+((?:(?!@param|@warning|@note|@return|@returns|@see|@code|@details).)+)', detailed, re.DOTALL)
        see_also = re.findall(r'@see\s+((?:(?!@param|@warning|@note|@return|@returns|@see|@code|@details).)+)', detailed, re.DOTALL)
        code_blocks = re.findall(r'@code(?:cpp)?\s*([\s\S]*?)\s*@endcode', detailed)
        details = re.findall(r'@details\s+((?:(?!@param|@warning|@note|@return|@returns|@see|@code|@details).)+)', detailed, re.DOTALL)

        detailed_clean = re.sub(r"@param\s+\w+\s+(?:(?!@param|@warning|@note|@return|@returns|@see|@code|@details).)+", "", detailed, flags=re.DOTALL).strip()
        detailed_clean = re.sub(r'@warning\s+((?:(?!@param|@warning|@note|@return|@returns|@see|@code|@details).)+)',"",detailed_clean, flags=re.DOTALL).strip()
        detailed_clean = re.sub(r'@note\s+((?:(?!@param|@warning|@note|@return|@returns|@see|@code|@details).)+)',"",detailed_clean, flags=re.DOTALL).strip()
        detailed_clean = re.sub(r'@return(?:s)?\s+((?:(?!@param|@warning|@note|@return|@returns|@see|@code|@details).)+)',"",detailed_clean, flags=re.DOTALL).strip()
        detailed_clean = re.sub(r'@see\s+((?:(?!@param|@warning|@note|@return|@returns|@see|@code|@details).)+)',"",detailed_clean, flags=re.DOTALL).strip()
        detailed_clean = re.sub(r'@code(?:cpp)?\s*([\s\S]*?)\s*@endcode',"",detailed_clean, flags=re.DOTALL).strip()
        detailed_clean = re.sub(r'@details\s+((?:(?!@param|@warning|@note|@return|@returns|@see|@code|@details).)+)',"",detailed_clean, flags=re.DOTALL).strip()
        
        def clean_and_reflow(text):
            if not text: return ""
            text = re.sub(r'^\s*\*\s?', '', text, flags=re.MULTILINE).strip()
            text = re.sub(r'(?<!\n)\n(?![ \t]*(?:\n|-|\*|\d+\.\s))', ' ', text)
            text = re.sub(r'(?<!\n)(\n)([ \t]*\d+\.\s)', r'\1\1\2', text)
            text = re.sub(r' +', ' ', text).strip()
            return text

        detailed_clean = clean_and_reflow(detailed_clean)
        brief = re.sub(r'\s+', ' ', brief).strip()
        
        numpy_docstring = f"{brief.replace('*', '').strip()}\n\n"
        numpy_docstring += f"{detailed_clean}\n\n"

        if details:
            details[0] = details[0]
            for detail_text in details:
                detail_text_clean = clean_and_reflow(detail_text)
                numpy_docstring += f"{textwrap.dedent(detail_text_clean)}\n\n"

        if params:
            numpy_docstring += "Parameters\n----------\n"
            for param_name, param_desc in params:
                param_type = param_types_dict.get(param_name, "Unknown")
                
                param_desc_clean = clean_and_reflow(param_desc)
                lines = param_desc_clean.split('\n')
                indented_lines = []
                
                if len(lines) > 1:
                    for line in lines[1:]:
                        if line.strip():
                            indented_lines.append("    " + line.strip())
                        else:
                            indented_lines.append("") # Preserve empty line

                if indented_lines:
                    final_desc = lines[0] + "\n" + "\n".join(indented_lines)
                else:
                    final_desc = lines[0]

                numpy_docstring += f"{param_name} : {param_type}\n    {final_desc}\n"

        if returns:
            numpy_docstring += "\nReturns\n-------\n"
            for ret_desc in returns:
                ret_desc_clean = clean_and_reflow(ret_desc)
                numpy_docstring += f"{ret_desc_clean}\n"
        
        if warnings:
            numpy_docstring += "\nWarnings\n--------\n"
            for warning in warnings:
                warning_clean = clean_and_reflow(warning)
                numpy_docstring += f"{warning_clean}\n"
        
        if notes:
            numpy_docstring += "\nNotes\n-----\n"
            for note_text in notes:
                note_text_clean = clean_and_reflow(note_text)
                numpy_docstring += f"{note_text_clean}\n"
        
        if see_also:
            numpy_docstring += "\nSee Also\n--------\n"
            for see in see_also:
                see_clean = clean_and_reflow(see)
                numpy_docstring += f"{see_clean}\n"
        
        if code_blocks:
            numpy_docstring += "\nExamples\n--------\n"
            for code_text in code_blocks:
                clean_code_text = code_text.replace('*', '')
                numpy_docstring += ".. code-block:: cpp\n\n    "
                numpy_docstring += ''.join([line+'\n' for line in clean_code_text.split('\n ')])
                numpy_docstring += "\n"
        
        if is_global:
            numpy_docstring += "\nAttention\n---------\n"
            numpy_docstring += "This is a wrapped CuPy `RawKernel` object,\n"
            numpy_docstring += "you can launch it as follows:\n\n"
            numpy_docstring += ".. code-block:: python\n\n"
            numpy_docstring += "   blocks = ...  # Number of blocks\n"
            numpy_docstring += "   threads = ...  # Number of threads per block\n"
            numpy_docstring += "   shared_memory = ...  # Dynamic shared memory buffer size (if needed)\n"
            numpy_docstring += f"   args = (...)  # Provide the arguments\n"
            numpy_docstring += "   kernel(grid, block, args, shared_mem=shared_memory)\n\n"
            numpy_docstring += "If you are new to CuPy, please see the CuPy `documentation`_.\n\n"
            numpy_docstring += ".. _documentation: https://cupy.dev/\n"

        return numpy_docstring

    return None

def get_kernel(module, kernel_name, max_shared_memory_limit=False):
    """Get kernel wrapper.

    Parameters
    ----------
    module : cupy.RawModule
        A CuPy rawModule
    kernel_name : str
        Name of the kernel to retrieve and wrap.
    
    Return
    ------
        Wrapper to the RawKernel with a NumPy-style docstring.
    """
    kernel = module.get_function(kernel_name)
    if max_shared_memory_limit:
        import cupy as cp
        props = cp.cuda.runtime.getDeviceProperties(cp.cuda.Device().id)
        shared_memory_limit = props.get('sharedMemPerBlockOptin', props['sharedMemPerBlock'])
        kernel.max_dynamic_shared_size_bytes = shared_memory_limit
    docstring = convert_doxygen_to_numpy(module.code, kernel_name)
    return wrap_cupy_rawkernel(kernel, docstring)